'''
affine (geometric) transformations
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License
import numpy as np
from skimage.transform import resize, rotate


def rotation(image, angle=0):
    '''Rotate an image around its centre. Uses skimage.transform.rotate. Areas
    that are rotated beyond the image are filled in with black pixel values

    Args:
        image (np.array): image to be rotated
        angle (float): amount of rotation in degrees (Defaults to 0)

    Returns:
        image (np.array): rotated image
    '''
    if angle == 0:
        return image
    return np.clip(rotate(image, angle, order=3), 0, 1)


def x_shift(image, x_shift=0):
    '''Translate image horizontally

    Args:
        image (np.array): image to be shifted
        x_shift (float): shift proportion of the image in the range (-1, 1).
                         (Defaults to 0)

    Returns:
        image (np.array): shifted image
    '''
    if x_shift == 0:
        return image
    return _translate_image(image, x_shift, 0)


def y_shift(image, y_shift=0):
    '''Translate image vertically

    Args:
        image (np.array): image to be shifted
        y_shift (float): shift proportion of the image in the range (-1, 1).
                         (Defaults to 0)

    Returns:
        image (np.array): shifted image
    '''
    if y_shift == 0:
        return image
    return _translate_image(image, 0, y_shift)


def _translate_image(image, x_shift=0, y_shift=0):
    '''Translate image vertically

    Args:
        image (np.array): image to be shifted
        x_shift (float): shift proportion of the image in the range (-1, 1).
                         (Defaults to 0)
        y_shift (float): shift proportion of the image in the range (-1, 1).
                         (Defaults to 0)

    Returns:
        image (np.array): shifted image
    '''
    if x_shift == 0 and y_shift == 0:
        return image
    original_size = image.shape
    canvas = np.zeros(original_size, dtype=image.dtype)
    prop_x = int(original_size[1]*abs(x_shift))
    prop_y = int(original_size[0]*abs(y_shift))
    # make sure we dont go off the canvas
    if original_size[1] - prop_x < 1 or original_size[0] - prop_y < 1:
        return canvas
    if y_shift >= 0 and x_shift >= 0:
        canvas[prop_y:, prop_x:, :] = image[:original_size[0] -
                                            prop_y, :original_size[1]-prop_x, :]
    elif y_shift < 0 and x_shift >= 0:
        canvas[:original_size[0]-prop_y, prop_x:, :] = image[prop_y -
                                                             original_size[0]:, :original_size[1]-prop_x, :]
    elif y_shift >= 0 and x_shift < 0:
        canvas[prop_y:, :original_size[1]-prop_x,
               :] = image[:original_size[0]-prop_y, prop_x-original_size[1]:, :]
    elif y_shift < 0 and x_shift < 0:
        canvas[:original_size[0]-prop_y:, :original_size[1]-prop_x,
               :] = image[prop_y-original_size[0]:, prop_x-original_size[1]:, :]
    return np.clip(canvas, 0, 1)


def zoom_image(image, scale_factor=1):
    '''digital zoom of image

    Args:
        image (np.array): image to be zoomed
        scale_factor (float): the percentage to zoom in by (for square zoom only).
                              0.5  = 2x zoom out
                              1 = normal size
                              2 = 2x zoom in
                              (Defaults to 1)

    Returns:
        image (np.array): zoomed image
    '''
    if scale_factor == 1:
        return image
    original_size = image.shape
    new_size_x = int(original_size[0]/scale_factor)
    new_size_y = int(original_size[1]/scale_factor)
    if scale_factor > 1:
        start_point_x = int((original_size[0] - new_size_x)/2)
        start_point_y = int((original_size[1] - new_size_y)/2)
        image = image[start_point_x:start_point_x+new_size_x,
                      start_point_y:start_point_y+new_size_y,
                      :]
    elif scale_factor < 1:
        new_size = (new_size_x, new_size_y, original_size[2]) if len(
            original_size) == 3 else (new_size_x, new_size_y)
        start_point_x = int((new_size_x - original_size[0])/2)
        start_point_y = int((new_size_y - original_size[1])/2)
        zoomed_out = np.zeros(new_size, dtype=image.dtype)
        zoomed_out[start_point_x:start_point_x+original_size[0],
                   start_point_y:start_point_y+original_size[1],
                   :] = image
        image = zoomed_out

    return np.clip(resize(image, original_size), 0, 1)



'''
create experiment window
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
import random
import threading
import warnings
import time

import numpy as np
import pandas as pd
from PyQt6.QtWidgets import (QMainWindow,
                             QHBoxLayout,
                             QVBoxLayout,
                             QTabWidget,
                             QApplication,
                             QPushButton,
                             QLabel,
                             QMessageBox)

from PyQt6.QtCore import Qt, pyqtSignal, pyqtSlot, QObject, QThread
from PyQt6.QtGui import QShortcut, QKeySequence

import IQM_Vis
from IQM_Vis.UI.custom_widgets import ClickLabel
from IQM_Vis.UI import utils
from IQM_Vis.utils import gui_utils, plot_utils, image_utils, save_utils


class make_experiment_2AFC(QMainWindow):
    saved_experiment = pyqtSignal(str)
    reset_clicked_image = pyqtSignal(dict)

    def __init__(self, 
                 checked_transformation_params, 
                 data_store, 
                 image_display_size,
                 rgb_brightness,
                 display_brightness,
                 default_save_dir=save_utils.DEFAULT_SAVE_DIR,
                 image_preprocessing='None',
                 image_postprocessing='None',
                 checked_metrics={}):
        super().__init__()
        self.checked_transformation_params = checked_transformation_params
        if self.checked_transformation_params == {}:
            return
        self.checked_metrics = checked_metrics
        self.data_store = data_store
        self.image_display_size = image_display_size
        self.rgb_brightness = rgb_brightness
        self.display_brightness = display_brightness
        self.default_save_dir = default_save_dir
        self.default_save_dir = os.path.join(default_save_dir, '2AFC')
        
        
        self.processing = {'pre': image_preprocessing,
                           'post': image_postprocessing}

        self.clicked_event = threading.Event()
        self.able_to_click = False

        self.image_change_worker = reset_image_widget_to_black()
        self.image_change_worker.completed.connect(self.click_completed)
        self.image_worker_thread = QThread()
        self.reset_clicked_image.connect(self.image_change_worker.change_to_solid)
        self.image_change_worker.moveToThread(self.image_worker_thread)
        self.image_worker_thread.start()

        self.stop_event = threading.Event()
        self.saved = False
        self.quit_experiment = False
        self.get_all_images()
        
        # make the experiment directory
        self.default_save_dir = os.path.join(
            self.default_save_dir, self.image_name)
        # make unique save dir name
        self.new_save_dir = self.get_unique_save_dir()
        
        self._init_experiment_window_widgets()
        self.get_metric_scores()
        self.experiment_layout()
        self.setCentralWidget(self.experiments_tab)
        self.setWindowTitle('2AFC Experiment')
        # move to centre of the screen
        qr = self.frameGeometry()
        cp = self.screen().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())
        self.show_all_images()

    def closeEvent(self, event):
        # Ask for confirmation if not saved
        if not self.saved:
            answer = QMessageBox.question(self,
            "Confirm Exit...",
            "Are you sure you want to exit?\nAll unsaved data will be lost.",
            QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes, 
                                          QMessageBox.StandardButton.Yes)
        else:
            answer = QMessageBox.StandardButton.Yes

        event.ignore()
        if answer == QMessageBox.StandardButton.Yes:
            self.quit_experiment = True
            if hasattr(self, 'range_worker'):
                self.range_worker.stop()
            if hasattr(self, 'image_worker_thread'):
                if self.image_worker_thread.isRunning():
                    self.image_worker_thread.quit()
                    self.image_worker_thread.wait()
                self.image_change_worker.stop()
            self.stop_event.set()
            self.clicked_event.set()
            event.accept()

    def quit(self):
        self.close()

    def show_all_images(self, tab='setup'):
        self.widget_experiments[tab]['images'].axes.axis('off')
        rows = int(len(self.experiment_transforms)**0.5)
        cols = int(np.ceil(len(self.experiment_transforms)/rows))
        for i, trans in enumerate(self.experiment_transforms):
            ax = self.widget_experiments[tab]['images'].figure.add_subplot(
                rows, cols, i+1)
            ax.imshow(image_utils.calibrate_brightness(
                trans['image'], self.rgb_brightness, self.display_brightness, ubyte=False))
            ax.axis('off')
            ax.set_title(save_utils.make_name_for_trans(trans), fontsize=6)
        # self.widget_experiments[tab]['images'].figure.tight_layout()

        # time.sleep(5)
        # QApplication.processEvents()
            
    def get_all_images(self):
        ''' save image name '''
        self.image_name = self.data_store.get_reference_image_name()
        ''' load all transformed images and sort them via MSE '''
        # get all the transform values
        self.experiment_trans_params = plot_utils.get_all_single_transform_params(
            self.checked_transformation_params, num_steps='from_dict')

        # remove any params with value 0 
        self.experiment_trans_params = [
            x for x in self.experiment_trans_params if not x[list(x.keys())[0]] == 0]

        # save the experiment ordering before reordering (for saving to csv col ordering)
        self.original_params_order = []
        for single_trans in self.experiment_trans_params:
            trans_name = list(single_trans.keys())[0]
            param = single_trans[trans_name]
            data = {'transform_name': trans_name,
                    'transform_value': param}
            self.original_params_order.append(
                save_utils.make_name_for_trans(data))

        # REFERENCE image
        self.ref_image = self.data_store.get_reference_image()
        if hasattr(self.data_store, 'get_reference_unprocessed'):
            self.ref_image_unprocessed = self.data_store.get_reference_unprocessed()
        # get MSE for experiments to get a rough sorting
        mses = []
        mse = IQM_Vis.metrics.MSE()
        for trans in self.experiment_trans_params:
            mses.append(
                mse(self.ref_image, self.get_single_transform_im(trans)))
        # put median MSE at the end (best for quick sort)
        self.experiment_trans_params = sort_list(
            self.experiment_trans_params, mses)  # sort array
        # take median out and get random shuffle for the rest
        median = self.experiment_trans_params.pop(
            len(self.experiment_trans_params)//2)
        random.shuffle(self.experiment_trans_params)
        self.experiment_trans_params.append(median)
        # load all images
        self.experiment_transforms = []
        for single_trans in self.experiment_trans_params:
            trans_name = list(single_trans.keys())[0]
            param = single_trans[trans_name]
            img = self.get_single_transform_im(single_trans)
            data = {'transform_name': trans_name,
                    'transform_value': param,
                    'image': img}
            self.experiment_transforms.append(data)
        self.calc_max_comparisons(num_images=len(self.experiment_transforms))
    
    def calc_max_comparisons(self, num_images):
        # calc expected number of comparisons - 
        # http://homepages.math.uic.edu/~leon/cs-mcs401-r07/handouts/quicksort-continued.pdf
        self.min_expected_comps = num_images * np.log(num_images)
        self.max_expected_comps = 1.39 * num_images * np.log(num_images)
    
    def get_metric_scores(self):
        '''get IQM scores to save alongside the experiment for plotting/analysis purposes'''
        IQM_scores = {}
        for data in self.experiment_transforms:
            score_dict = self.data_store.get_metrics(transformed_image=data['image'],
                                                 metrics_to_use=self.checked_metrics)
            scores = []
            metrics = []
            for name, score in score_dict.items():
                metrics.append(name)
                scores.append(float(score))
            IQM_scores[save_utils.make_name_for_trans(data)] = scores
        IQM_scores['IQM'] = metrics
        self.IQM_scores_df = pd.DataFrame.from_dict(IQM_scores)
        self.IQM_scores_df.set_index('IQM', inplace=True)

    def _init_experiment_window_widgets(self):
        self.widget_experiments = {'exp': {}, 'preamble': {}, 'setup': {}, 'final':{}}
        ''' setup tab '''
        self.widget_experiments['setup']['start_button'] = QPushButton(
            'Setup', self)
        self.widget_experiments['setup']['start_button'].clicked.connect(self.setup_experiment)
        self.widget_experiments['setup']['quit_button'] = QPushButton('Quit', self)
        self.widget_experiments['setup']['quit_button'].clicked.connect(self.quit)
        QShortcut(QKeySequence("Ctrl+Q"),
                  self.widget_experiments['setup']['quit_button'], self.quit)
        self.widget_experiments['setup']['images'] = gui_utils.MplCanvas(size=None)
        self.widget_experiments['setup']['text'] = QLabel(self)
        self.widget_experiments['setup']['text'].setText(f'''
        2AFC Experiment to be setup with the above images using the settings:
            Save folder: {self.default_save_dir}
            Image Display Size: {self.image_display_size}
            Image Calibration:
                Max RGB Brightness: {self.rgb_brightness}
                Max Display Brightness: {self.display_brightness}

            Expected Number of Comparisons: {int(self.min_expected_comps)}
            MAX Expected Number of Comparisons: {int(self.max_expected_comps)}

        Click the Setup button to setup up the experiment and hand over to the test subject.
        ''')
        # self.widget_experiments['setup']['text'].setAlignment(
        #     Qt.AlignmentFlag.AlignCenter)

        ''' info tab '''
        self.widget_experiments['preamble']['text'] = QLabel(self)
        self.widget_experiments['preamble']['text'].setText('''
        For this experiment you will be shown a reference image and two similar images.

        You need to click on the image (A or B) which you believe to be most similar to the reference image.




        When you are ready, click the Start button to begin the experiment ''')
        self.running_experiment = False
        self.widget_experiments['preamble']['start_button'] = QPushButton('Start', self)
        self.widget_experiments['preamble']['start_button'].clicked.connect(self.toggle_experiment)

        self.widget_experiments['preamble']['quit_button'] = QPushButton('Quit', self)
        self.widget_experiments['preamble']['quit_button'].clicked.connect(
            self.quit)
        QShortcut(QKeySequence("Ctrl+Q"),
                  self.widget_experiments['preamble']['quit_button'], self.quit)

        ''' experiment tab '''
        self.exp_info_text = 'Click on which image, A or B, is most similar to the reference image (or use left/right arrows)'
        self.widget_experiments['exp']['info'] = QLabel(
            self.exp_info_text, self)
        for image in ['Reference', 'A', 'B']:
            self.widget_experiments['exp'][image] = {}
            self.widget_experiments['exp'][image]['data'] = ClickLabel(image)
            self.widget_experiments['exp'][image]['data'].setAlignment(Qt.AlignmentFlag.AlignCenter)
            # image label
            self.widget_experiments['exp'][image]['label'] = QLabel(image, self)
            self.widget_experiments['exp'][image]['label'].setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.widget_experiments['exp']['A']['data'].clicked.connect(self.clicked_image)
        self.widget_experiments['exp']['B']['data'].clicked.connect(self.clicked_image)
        # shortcuts keyboard
        QShortcut(QKeySequence(
            "left"), self.widget_experiments['exp']['A']['data'], self.widget_experiments['exp']['A']['data'].mousePressEvent)
        QShortcut(QKeySequence(
            "right"), self.widget_experiments['exp']['B']['data'], self.widget_experiments['exp']['B']['data'].mousePressEvent)
        # quit
        self.widget_experiments['exp']['quit_button'] = QPushButton('Quit', self)
        self.widget_experiments['exp']['quit_button'].clicked.connect(self.quit)
        QShortcut(QKeySequence("Ctrl+Q"),
                  self.widget_experiments['exp']['quit_button'], self.quit)

        ''' finish tab '''
        self.widget_experiments['final']['order_text'] = QLabel(
            'Experiment Sorting Order:', self)
        self.widget_experiments['final']['images'] = gui_utils.MplCanvas(size=None)
        self.widget_experiments['final']['quit_button'] = QPushButton('Quit', self)
        self.widget_experiments['final']['quit_button'].clicked.connect(
            self.quit)
        QShortcut(QKeySequence("Ctrl+Q"),
                self.widget_experiments['final']['quit_button'], self.quit)
        self.widget_experiments['final']['save_label'] = QLabel('Not saved yet', self)

    def experiment_layout(self):
        ''' setup '''
        experiment_text = QVBoxLayout()
        experiment_text.addWidget(self.widget_experiments['setup']['text'])
        experiment_setup_buttons = QHBoxLayout()
        experiment_setup_buttons.addWidget(
            self.widget_experiments['setup']['start_button'])
        experiment_setup_buttons.addWidget(
            self.widget_experiments['setup']['quit_button'])
        experiment_text.addLayout(experiment_setup_buttons)

        experiment_mode_setup = QVBoxLayout()
        experiment_mode_setup.addWidget(self.widget_experiments['setup']['images'])
        experiment_mode_setup.addLayout(experiment_text)
        experiment_mode_setup.setAlignment(Qt.AlignmentFlag.AlignCenter)
        experiment_mode_setup.addStretch()


        ''' info '''
        experiment_info_buttons = QHBoxLayout()
        experiment_info_buttons.addWidget(
            self.widget_experiments['preamble']['start_button'])
        experiment_info_buttons.addWidget(
            self.widget_experiments['preamble']['quit_button'])

        experiment_mode_info = QVBoxLayout()
        experiment_mode_info.addWidget(
            self.widget_experiments['preamble']['text'])
        experiment_mode_info.setAlignment(Qt.AlignmentFlag.AlignCenter)
        experiment_mode_info.addLayout(experiment_info_buttons)

        ''' experiment '''
        info = QVBoxLayout()
        info.addWidget(self.widget_experiments['exp']['info'])
        info.setAlignment(Qt.AlignmentFlag.AlignCenter)

        quit = QVBoxLayout()
        quit.addWidget(self.widget_experiments['exp']['quit_button'])
        quit.setAlignment(Qt.AlignmentFlag.AlignCenter)

        layouts = []
        for im in ['A', 'Reference', 'B']:
            _layout = QVBoxLayout()
            for _, widget in self.widget_experiments['exp'][im].items():
                _layout.addWidget(widget)
            _layout.setAlignment(Qt.AlignmentFlag.AlignTop)
            layouts.append(_layout)

        # add images to h box
        experiment_images = QHBoxLayout()
        for layout in layouts:
            experiment_images.addLayout(layout)
        experiment_images.setAlignment(Qt.AlignmentFlag.AlignTop)

        run_experiment = QVBoxLayout()
        run_experiment.addLayout(info)
        run_experiment.addLayout(experiment_images)
        run_experiment.addLayout(quit)
        run_experiment.setAlignment(Qt.AlignmentFlag.AlignCenter)

        ''' finished '''
        finish_experiment = QVBoxLayout()
        finish_experiment.addWidget(self.widget_experiments['final']['order_text'])
        finish_experiment.addWidget(self.widget_experiments['final']['images'])
        finish_experiment.addWidget(self.widget_experiments['final']['save_label'])
        finish_experiment.addWidget(self.widget_experiments['final']['quit_button'])
        finish_experiment.setAlignment(Qt.AlignmentFlag.AlignCenter)
        finish_experiment.addStretch()

        self.experiments_tab = QTabWidget()
        for tab_layout, tab_name in zip([experiment_mode_setup, experiment_mode_info, run_experiment, finish_experiment],
                                        ['setup', 'info', 'run', 'finish']):
            utils.add_layout_to_tab(self.experiments_tab, tab_layout, tab_name)
        # experiment_mode_layout = QVBoxLayout()
        # experiment_mode_layout.addWidget(self.experiments_tab)
        # return experiment_mode_layout

    ''' experiment running functions'''
    def setup_experiment(self):
        self.click_counter = 0
        self.max_text = f'{int(self.min_expected_comps)}-{int(self.max_expected_comps)}'
        self.widget_experiments['exp']['info'].setText(
            f'{self.exp_info_text} {self.click_counter}/{self.max_text}')
        self.experiments_tab.setCurrentIndex(1)
        self.experiments_tab.setTabEnabled(0, False)
        self.experiments_tab.setTabEnabled(2, False)
        self.experiments_tab.setTabEnabled(3, False)

    def toggle_experiment(self):
        if self.running_experiment:
            self.reset_experiment()
            self.experiments_tab.setTabEnabled(0, True)
            self.experiments_tab.setTabEnabled(1, True)
            self.running_experiment = False
            # self.widget_experiments['preamble']['start_button'].setText('Start')

        else:
            self.experiments_tab.setTabEnabled(2, True)
            self.start_experiment()
            self.experiments_tab.setTabEnabled(0, False)
            self.experiments_tab.setTabEnabled(1, False)
            # self.widget_experiments['preamble']['start_button'].setText('Reset')
            self.running_experiment = True

    def reset_experiment(self):
        self.experiments_tab.setCurrentIndex(1)
        self.init_style('light')

    def start_experiment(self):
        self.running_experiment = True
        self.init_style('dark')
        self.experiments_tab.setCurrentIndex(2)

        # Display reference image
        gui_utils.change_im(self.widget_experiments['exp']['Reference']['data'], self.ref_image,
                            resize=self.image_display_size, rgb_brightness=self.rgb_brightness, display_brightness=self.display_brightness)

        # get user sorting
        self.sorting_thread = threading.Thread(target=self.quick_sort)
        self.sorting_thread.start()
        # self.quick_sort(0, len(self.experiment_transforms)-1)

    def finish_experiment(self):
        self.running_experiment = False
        self.experiments_tab.setTabEnabled(3, True)
        self.show_all_images(tab='final')
        self.init_style('light')
        self.experiments_tab.setCurrentIndex(3)
        self.experiments_tab.setTabEnabled(2, False)
        self.save_experiment()
        if self.saved == True:
            self.widget_experiments['final']['save_label'].setText(f'Saved to {self.default_save_dir}')
        else:
            self.widget_experiments['final']['save_label'].setText(f'Save failed to {self.default_save_dir}')

    def get_trans_funcs(self):
        # get the current transform functions
        trans_funcs = {}
        for single_trans in self.experiment_trans_params:
            trans_name = list(single_trans.keys())[0]
            trans_funcs[trans_name] = self.checked_transformation_params[trans_name]['function']
        return trans_funcs
    
    def get_unique_save_dir(self):
        # get the current transform functions
        trans_funcs = self.get_trans_funcs()

        # get a unique directory (same image with diff trans need a new dir)
        i = 1
        unique_dir_found = False
        new_dir = True
        while unique_dir_found == False:
            exp_save_dir = os.path.join(
                self.default_save_dir, f'experiment-{i}')
            if os.path.exists(exp_save_dir):
                # get transform funcs and params
                exp_trans_params = save_utils.load_obj(
                    os.path.join(exp_save_dir, 'transforms', 'transform_params.pkl'))
                exp_trans_funcs = save_utils.load_obj(
                    os.path.join(exp_save_dir, 'transforms', 'transform_functions.pkl'))

                # get image processing saved params
                processing_file = save_utils.get_image_processing_file(
                    exp_save_dir)
                procesing_same = False
                if os.path.exists(processing_file):
                    processing = save_utils.load_json_dict(processing_file)
                    if processing == self.processing:
                        procesing_same = True

                # check if experiment is the same
                if ((exp_trans_params == self.original_params_order) 
                    and (trans_funcs == exp_trans_funcs) 
                    and procesing_same):
                    self.default_save_dir = exp_save_dir
                    unique_dir_found = True
                    new_dir = False
                else:
                    i += 1
            else:
                self.default_save_dir = exp_save_dir
                unique_dir_found = True
        return new_dir
    
    def save_experiment(self):
        # get the current transform functions
        trans_funcs = self.get_trans_funcs()

        # make all the dirs and subdirs
        os.makedirs(self.default_save_dir, exist_ok=True)
        os.makedirs(os.path.join(self.default_save_dir, 'images'), exist_ok=True)
        os.makedirs(os.path.join(self.default_save_dir, 'transforms'), exist_ok=True)

        # save experiment images
        if not os.path.exists(save_utils.get_original_image_file(self.default_save_dir)):
            image_utils.save_image(self.ref_image,
                                    save_utils.get_original_image_file(self.default_save_dir))
        if not os.path.exists(save_utils.get_original_unprocessed_image_file(self.default_save_dir)):
            if hasattr(self, 'ref_image_unprocessed'):
                image_utils.save_image(self.ref_image_unprocessed,
                                        save_utils.get_original_unprocessed_image_file(self.default_save_dir))
        if self.new_save_dir == True:
            for trans in self.experiment_transforms:
                image_utils.save_image(
                    trans['image'], os.path.join(self.default_save_dir, 'images', f'{save_utils.make_name_for_trans(trans)}.png'))
            # save the transformations
            save_utils.save_obj(
                save_utils.get_transform_params_file(self.default_save_dir),
                self.original_params_order)
            save_utils.save_obj(
                save_utils.get_transform_functions_file(self.default_save_dir),
                dict(sorted(trans_funcs.items())))
            # save the image pre/post processing options
            save_utils.save_json_dict(
                save_utils.get_image_processing_file(self.default_save_dir),
                self.processing)

        # save the experiment results
        exp_order = []
        for trans in self.experiment_transforms:
            exp_order.append(save_utils.make_name_for_trans(trans))
        csv_file = save_utils.save_2AFC_experiment_results(
            self.original_params_order,
            exp_order,
            self.default_save_dir,
            self.times_taken,
            self.IQM_scores_df)
        self.saved = True
        self.saved_experiment.emit(csv_file)

    ''' sorting algorithm resource: https://www.geeksforgeeks.org/quick-sort/'''
    def quick_sort(self):
        self.times_taken = []
        self._quick_sort(0, len(self.experiment_transforms)-1)
        if self.quit_experiment != True:
            self.finish_experiment()
            # for trans in self.experiment_transforms:
            #     print(trans['transform_name'], trans['transform_value'])

    def _quick_sort(self, low, high):
        if low < high:
            # Find pivot elements such that element smaller than pivot are on the
            # left and elements greater than pivot are on the right
            pi = self.partition(low, high)
            if self.stop_event.is_set():
                return
            # Recursive call on the left of pivot
            self._quick_sort(low, pi - 1)
            # Recursive call on the right of pivot
            self._quick_sort(pi + 1, high)

    def partition(self, low, high):
        ''' given an unsorted partition of the array between low and high, order
            elements lower than a given pivot point to the left and higher to the right'''
        # Choose the end element as pivot
        self.high = high
        self.low = low
        self.current_comparision = low
        self.pivot = self.experiment_transforms[self.high]
        # Pointer for greater element
        self.comp_pointer = low - 1
        # Traverse through all elements and compare each element with pivot (by user clicking)
        while True:
            time0 = time.time()
            # randomly assign to image A or B
            ims_to_display = [
                self.experiment_transforms[self.current_comparision], self.pivot]
            random.shuffle(ims_to_display)
            # display the images
            self.change_experiment_images(A_trans=ims_to_display[0],
                                          B_trans=ims_to_display[1])
            # wait for image to be clicked
            self.clicked_event.clear()
            self.able_to_click = True
            self.clicked_event.wait()
            if self.stop_event.is_set():
                return
            if self.less_than_pivot == True:
            # if self.experiment_transforms[self.current_comparision]['brightness'] <= self.pivot['brightness']:
                # If element smaller than pivot is found swap it with the greater element pointed by i
                self.comp_pointer += 1
                # Swapping element at i with element at j
                self.swap_inds(self.comp_pointer, self.current_comparision)
            self.current_comparision += 1
            self.times_taken.append(time.time()-time0)
            if self.current_comparision == self.high:
                break
        # Swap the pivot element with the greater element specified by i
        self.swap_inds(self.comp_pointer+1, self.high)
        # Return the position from where partition is done
        return self.comp_pointer + 1

    def clicked_image(self, image_name, widget_name):
        if self.able_to_click == False:
            return
        if self.running_experiment == False:
            return
        self.click_counter += 1

        self.widget_experiments['exp']['info'].setText(f'{self.exp_info_text} {self.click_counter+1}/{self.max_text}')
        
        self.able_to_click = False
        # get comparison to pivot
        trans_str = image_name[len(self.image_name)+1:]
        if trans_str != save_utils.make_name_for_trans(self.pivot): # lower value
            # If element smaller than pivot is found swap it with the greater element pointed by i
            self.less_than_pivot = True
        else:
            self.less_than_pivot = False
        # make clicked image black to show user
        data = {'image_display_size': self.image_display_size,
                'widget': self.widget_experiments['exp'][widget_name]['data']}
        self.reset_clicked_image.emit(data)  # change to black image, after x amount of time will change to experimetn image
        
    def click_completed(self):
        # unlock the wait
        self.clicked_event.set()

    def swap_inds(self, i, j):
        (self.experiment_transforms[i], self.experiment_transforms[j]) = (self.experiment_transforms[j], self.experiment_transforms[i])

    def get_single_transform_im(self, single_trans_dict):
        trans_name = list(single_trans_dict)[0]
        return image_utils.get_transform_image(self.data_store,
                                        {trans_name: self.checked_transformation_params[trans_name]},
                                        single_trans_dict)

    def change_experiment_images(self, A_trans, B_trans):
        A = A_trans['image']
        B = B_trans['image']
        
        gui_utils.change_im(self.widget_experiments['exp']['A']['data'], A, resize=self.image_display_size,
                            rgb_brightness=self.rgb_brightness, display_brightness=self.display_brightness)
        self.widget_experiments['exp']['A']['data'].setObjectName(
            f'{self.image_name}-{save_utils.make_name_for_trans(A_trans)}')
        gui_utils.change_im(self.widget_experiments['exp']['B']['data'], B, resize=self.image_display_size,
                            rgb_brightness=self.rgb_brightness, display_brightness=self.display_brightness)
        self.widget_experiments['exp']['B']['data'].setObjectName(
            f'{self.image_name}-{save_utils.make_name_for_trans(B_trans)}')

    ''' UI '''
    def init_style(self, style='light', css_file=None):
        if css_file == None:
            dir = os.path.dirname(os.path.abspath(__file__))
            # css_file = os.path.join(dir, 'style-light.css')
            css_file = os.path.join(dir, f'style-{style}.css')
        if os.path.isfile(css_file):
            with open(css_file, 'r') as file:
                self.setStyleSheet(file.read())
        else:
            warnings.warn('Cannot load css style sheet - file not found')


class reset_image_widget_to_black(QObject):
    ''' change clicked image to black and pause '''
    completed = pyqtSignal(float)

    def __init__(self, time=0.1):
        super().__init__()
        self.running = True
        self.time = time
        self.black_array = np.zeros([100, 100, 3])
        # self.black_array[:, :, 1] = 1 # blue

    @pyqtSlot(dict)
    def change_to_solid(self, data):
        t_start = time.time()
        image_display_size = data['image_display_size']
        widget = data['widget']
        # make clicked image black to show user
        gui_utils.change_im(widget,
                            self.black_array,
                            resize=image_display_size)
        # pause for half the time needed (will use the loop below to wait for full time and also see if image has turned black yet)
        time.sleep(self.time/2)
        m = 1
        t_time = time.time() - t_start
        # wait until black image is shown
        while (not (m == 0 or m == 63.75)) and (t_time < self.time): # 63.75 is when 4th channel is all ones , rest are 0
            time.sleep(self.time/10)
            # get the image data from the widget to check if it's been made black yet
            pixmap = widget.pixmap()
            q_img = pixmap.toImage()
            ptr = q_img.bits()
            ptr = q_img.constBits()
            ptr.setsize(q_img.sizeInBytes())
            np_img = np.array(ptr, copy=False).reshape(
                q_img.height(), q_img.width(), 4)
            m = np_img.mean()
            # calc if the time is up yet
            t_time = time.time() - t_start
        # all complete so send signal 
        self.completed.emit(1.0)

    def stop(self):
        self.running = False

    def __del__(self):
        # close app upon garbage collection
        self.stop()


def sort_list(list1, list2):
    # sort list1 based on list2
    inds = np.argsort(list2)
    sorted_list1 = []
    for i in inds:
        sorted_list1.append(list1[i])
    return sorted_list1



''' KODAK dataset '''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import IQM_Vis
from IQM_Vis.examples.KODAK_dataset import KODAK_IMAGES

def run():
    metrs = IQM_Vis.metrics.get_all_metrics()
    if '1-MS_SSIM' in metrs:
      metrs.pop('1-MS_SSIM')
    data = IQM_Vis.dataset_holder(KODAK_IMAGES,
                                  metrs,
                                #   IQM_Vis.metrics.get_all_metric_images()
                                  )


    IQM_Vis.make_UI(data,
                    IQM_Vis.transforms.get_all_transforms(),
                    restrict_options=3
                    )
    IQM_Vis.make_UI()

if __name__ == '__main__':
    run()



'''
Utils for PyQt6 image, text and graph widgets
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

try:
    from PyQt6.QtGui import QPixmap, QImage
    import matplotlib; matplotlib.use('Qt5Agg')
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
    from matplotlib.figure import Figure
    import matplotlib.pyplot as plt
    HEADLESS = False
except ImportError:
    import warnings
    warnings.warn('Can not load PyQt6 library - running IQM_Vis package in headless mode')
    HEADLESS = True
import pickle
import io

import numpy as np
from skimage.util import img_as_ubyte
from skimage.transform import resize
from IQM_Vis.utils.image_utils import resize_to_longest_side, calibrate_brightness

'''
image helper functions
'''
# def resize_im_to(np_array, size):
#     down_im = resize(np_array, size)
#     return img_as_ubyte(down_im)


def change_im(widget, im, resize=False, rgb_brightness=250, display_brightness=250, border=False):
    '''
    given a numpy image, changes the given widget Frame
    '''
    # clip image to safe bounds
    im = np.clip(im, 0, 1)
    # convert to three channel
    if len(im.shape) == 2:
        im = im[..., np.newaxis]
    if im.shape[2] == 1:
        im = np.concatenate([im, im, im], axis=2)
    if resize:
        im = resize_to_longest_side(im, resize)
        im = img_as_ubyte(im)
    im = calibrate_brightness(im, rgb_brightness, display_brightness)
    if border == True:
        # add a black border to the image
        im = np.pad(im, ((5, 5), (5, 5), (0, 0)), 'constant', constant_values=0)
        if resize:
            im = resize_to_longest_side(im, resize)
            im = img_as_ubyte(im)
    qimage = QImage(im,
                    im.shape[1],
                    im.shape[0],
                    im.shape[1]*im.shape[2],
                    QImage.Format.Format_RGB888)
    pixmap = QPixmap(qimage)
    widget.setPixmap(pixmap)

'''
text utils
'''

def str_to_len(string, length=5, append_char='0', plus=False):
    # cut string to length, or append character to make to length
    if string[0] !=  '-' and plus == True:
        string = '+' + string
    if len(string) > length:
        string = string[:length]
    elif len(string) < length:
        string = string + append_char*(length-len(string))
    return string

def get_metric_image_name(metric, data_store):
    return metric+get_image_pair_name(data_store)

def get_trans_dict_from_str(trans_str, return_dict=False):
    # determine which experiment formatting we are using
    if len(trans_str.split('-----')) != 1:
        splitter = '-----'   # legacy code experiment format - illegible with negative numbers
    elif len(trans_str.split('::')) != 1:
        splitter = '::'   # legacy code experiment format - didn't work with windows filesystem
    elif len(trans_str.split('<>')) != 1:
        splitter = '<>'   # newer code experiment format
    elif len(trans_str.split('=')) != 1:
        splitter = '='   # legacy experiment format
    else:
        raise ValueError('No splitter found in transformation string')
    # split the string into the transformation and the value    
    trans = splitter.join(trans_str.split(splitter)[:-1])
    # error with duplicated columns in csv with pandas load
    try:
        trans_value = float(trans_str.split(splitter)[-1])   # value is last part of string
    except ValueError:
        error_value = trans_str.split(splitter)[-1]
        # remove the '.1' etc appended by pandas
        value = '.'.join(error_value.split('.')[:-1])
        trans_value = float(value)
    if return_dict == True:
        return {trans: trans_value}
    else:
        return trans, trans_value

def get_transformed_image_name(data_store):
    return 'T('+data_store.get_image_to_transform_name()+')'

def get_image_pair_name(data_store):
    return str((data_store.get_reference_image_name(), get_transformed_image_name(data_store)))

def get_resolutions(data_store):
    ref = data_store.get_reference_image().shape
    if hasattr(data_store, 'image_post_processing'):
        trans_im = data_store.get_image_to_transform()
        if data_store.image_post_processing is not None:
            trans_im = data_store.image_post_processing(trans_im)
    trans = trans_im.shape
    return {'reference': f'{ref[0]}x{ref[1]}',
            'transform': f'{trans[0]}x{trans[1]}'}


'''
matplotlib widget utils
'''
if not HEADLESS:
    # Get a matplotlib canvas as a Qt Widget
    class MplCanvas(FigureCanvasQTAgg):
        def __init__(self, size=(3.5, 3.5), dpi=100, polar=False):
            self.figure = Figure(figsize=size)  # , dpi=dpi)
            self.axes = self.figure.add_subplot(111, polar=polar)
            super(MplCanvas, self).__init__(self.figure)



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License
import numpy as np

# from torchmetrics import PeakSignalNoiseRatio as PSNRs
from IQM_Vis.metrics.metric_utils import _check_shapes

class MAE:
    '''Mean Absolute Error between two images. Images must have the same
       dimensions

    Args:
        return_image (bool): Whether to return the image (Defaults to False which
                             will return a scalar value)
    '''

    def __init__(self, return_image=False):
        self.return_image = return_image

    def __call__(self, im_ref, im_comp, **kwargs):
        '''When an instance is called

        Args:
            im_ref (np.array): Reference image
            im_comp (np.array): Comparison image
            **kwargs: Arbitrary keyword arguments

        Returns:
            score (np.array): MAE (scalar if return_image is False, image if
                             return_image is True)
        '''
        _check_shapes(im_ref, im_comp)
        L1 = np.abs(im_ref - im_comp)
        if self.return_image:
            return L1
        else:
            return L1.mean()


class MSE:
    '''Mean Squared Error between two images. Images must have the same
       dimensions

    Args:
        return_image (bool): Whether to return the image (Defaults to False which
                             will return a scalar value)
    '''

    def __init__(self, return_image=False):
        self.return_image = return_image

    def __call__(self, im_ref, im_comp, **kwargs):
        '''When an instance is called

        Args:
            im_ref (np.array): Reference image
            im_comp (np.array): Comparison image
            **kwargs: Arbitrary keyword arguments

        Returns:
            score (np.array): MSE (scalar if return_image is False, image if
                             return_image is True)
        '''
        _check_shapes(im_ref, im_comp)
        L2 = np.square(im_ref - im_comp)
        if self.return_image:
            return L2
        else:
            return L2.mean()


class RMSE:
    '''Root Mean Squared Error between two images. Images must have the same
       dimensions

    Args:
        return_image (bool): Whether to return the image (Defaults to False which
                             will return a scalar value)
    '''

    def __init__(self, return_image=False):
        self.return_image = return_image

    def __call__(self, im_ref, im_comp, **kwargs):
        '''When an instance is called

        Args:
            im_ref (np.array): Reference image
            im_comp (np.array): Comparison image
            **kwargs: Arbitrary keyword arguments

        Returns:
            score (np.array): RMSE (scalar if return_image is False, image if
                             return_image is True)
        '''
        _check_shapes(im_ref, im_comp)
        L2 = np.square(im_ref - im_comp)
        if self.return_image:
            return np.sqrt(L2)
        else:
            return np.sqrt(L2.mean())


class one_over_PSNR:
    '''Peak signal to noise ratio - https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio
    The score given is normalised between 0, 1 by taking 1/PSNR 

    Args:
        return_image (bool): Whether to return the image (Defaults to False which
                             will return a scalar value)
    '''

    def __init__(self):
        self.initialised = False
        # self.metric = PSNRs
        # self.device = torch.device(
        #     "cuda" if torch.cuda.is_available() else "cpu")
        # self.preproccess_function = _numpy_to_torch_image

    def __call__(self, im_ref, im_comp, **kwargs):
        '''When an instance is called

        Args:
            im_ref (np.array): Reference image
            im_comp (np.array): Comparison image
            **kwargs: Arbitrary keyword arguments

        Returns:
            score (np.array): PSNR between the images
        '''
        _check_shapes(im_ref, im_comp)
        se = np.square(im_ref - im_comp)
        mse = se.mean()
        if mse == 0:
            return 0
        psnr = 10 * np.log10(1/(mse))
        return 1/psnr

        # im_ref = self.preproccess_function(im_ref).to(
        #     device=self.device, dtype=torch.float)
        # im_comp = self.preproccess_function(im_comp).to(
        #     device=self.device, dtype=torch.float)
        # # set up metric
        # if self.initialised == False:
        #     with warnings.catch_warnings():    # we don't care about the warnings these give
        #         warnings.simplefilter("ignore")
        #         self._metric = self.metric()
        #         self._metric.to(self.device)
        # _score = self._metric(im_ref*255, im_comp*255).cpu().detach().numpy()
        # _score = 1 - _score
        # self._metric.reset()
        # return _score/255



import os

file_path = os.path.dirname(os.path.abspath(__file__))

IMAGE1 = os.path.join(file_path, 'waves1.jpeg')
IMAGE2 = os.path.join(file_path, 'waves2.jpeg')
IMAGE3 = os.path.join(file_path, 'waves3.jpeg')

DEFAULT_IMAGES = [IMAGE1, IMAGE2, IMAGE3]



'''
noise transformations
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from abc import ABC, abstractmethod
import numpy as np
import warnings

# Suppress the specific DeprecationWarning
warnings.filterwarnings("ignore", category=DeprecationWarning,
                        message="__array_wrap__ must accept context and return_scalar arguments")


class _base_noise(ABC):
    def __init__(self, acceptable_percent=0.9, max_iter=50, reject_low_noise=False, **kwargs):
        self._setup_args(**kwargs)
        self.acceptable_percent = acceptable_percent
        self.max_iter = max_iter
        self.reject_low_noise = reject_low_noise
        self.num_rejected = 0

    def _setup_args(self, **kwargs):
        pass

    @abstractmethod
    def _make_noisey_image(self, img, np=np):
        # uses any args on self for parameters
        # must return x_noisey, additive_noise
        pass

    def __call__(self, img, param=0):
        '''
        will return a noisy image with noise level according to 'param'
        highest noise will be returned if noise level is too low acccording to acceptable_percent
        '''
        self.num_rejected = 0
        self.actual_noise = 0
        if param is not None:
            self.param = param
        return self._call_single(img)

    def _call_single(self, img, np=np):
        '''
        make noisy image until either acceptable percent is reached or max_iter is reached
        will return the highest noise level if noise level is too low acccording to acceptable_percent
        '''
        for _ in range(self.max_iter):
            x_noisey, additive_noise = self._make_noisey_image(img)
            # check noise level and actual noise level reduced after clipping
            expected_noise = np.sqrt(np.mean(np.square(additive_noise)))
            if self.reject_low_noise == False:
                return x_noisey
            else:
                curr_actual_noise = np.sqrt(np.mean(np.square(x_noisey - img)))
                # option to reject and redo the noise if too low (recursion)
                if curr_actual_noise >= self.acceptable_percent * expected_noise:
                    return x_noisey
                elif curr_actual_noise > self.actual_noise:
                    self.num_rejected += 1
                    self.actual_noise = curr_actual_noise
                    best = x_noisey
                else:
                    self.num_rejected += 1
        return best


class noise_hypersphere(_base_noise):
    '''Add random noise on a hypersphere with radius epsilon. Image is clipped to the range (0, 1)

    Args:
        image (np.array): image to have noise added
        std (float): epsilon of the hypersphere radius of noise.
                         (Defaults to 0)

    Returns:
        image (np.array): adjusted image
    '''
    def _setup_args(self, epsilon=1, seed=True):
        self.param = epsilon
        self.seed = seed

    def _make_noisey_image(self, img):
        if self.param <= 0:
            return img, 0.0
        if self.seed:
            np.random.seed(42)
        noise = np.random.randn(*img.shape)
        noise_norm = noise / np.linalg.norm(noise.reshape(-1))
        additive_noise = self.param * noise_norm
        unclipped = img + additive_noise
        x_noisey = np.clip(unclipped, 0, 1)
        return x_noisey, additive_noise


class Gaussian_noise(_base_noise):
    '''Add Gaussian noise to image. Image is clipped to the range (0, 1)

    Args:
        image (np.array): image to have Gaussian noise added
        std (float): Standard deviation of the Gaussian noise.
                         (Defaults to 0)

    Returns:
        image (np.array): adjusted image
    '''
    def _setup_args(self, std=0, seed=True):
        self.param = std
        self.seed = seed

    def _make_noisey_image(self, img):
        if self.param <= 0:
            return img, 0.0
        if self.seed:
            np.random.seed(42)
        additive_noise = np.random.normal(
            loc=0, scale=self.param, size=img.shape)
        unclipped = img + additive_noise
        x_noisey = np.clip(unclipped, 0, 1)
        return x_noisey, additive_noise
    

def salt_and_pepper_noise(image, prob=0):
    '''
    Add salt and pepper noise to image

    Args: 
        image (np.array): image to be compressed
        prob: Probability of the noise (Defaults to 0).

    Returns:
        image (np.array): image with salt and pepper noise
    '''
    if prob == 0:
        return image
    min_val = 0
    max_val = 1
    image = image.copy()
    if len(image.shape) == 2:
        black = min_val
        white = max_val
    else:
        colorspace = image.shape[2]
        if colorspace == 3:  # RGB
            black = np.array([min_val, min_val, min_val], dtype='float')
            white = np.array([max_val, max_val, max_val], dtype='float')
        else:  # RGBA
            black = np.array(
                [min_val, min_val, min_val, max_val], dtype='float')
            white = np.array(
                [max_val, max_val, max_val, max_val], dtype='float')
    probs = np.random.random(image.shape[:2])
    image[probs < (prob / 2)] = black
    image[probs > 1 - (prob / 2)] = white
    return np.clip(image, 0, 1)


if __name__ == '__main__':
    import numpy as np
    image = np.random.rand(256, 256, 3)
    noise = Gaussian_noise()
    noise(image, 0)
    # noise(image, 0.1)



# tests/conftest.py

import pytest
from PyQt6.QtWidgets import QApplication
# import matplotlib

# # Use a different backend to avoid conflicts
# matplotlib.use('Agg')


@pytest.fixture(scope="session", autouse=True)
def qapp():
    """
    Ensure there's a QApplication instance running.
    This fixture is automatically used for each test session.
    """
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    yield app
    # Properly quit the application after the tests
    app.quit()



'''
Sample image transformations to get the user started with
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from skimage.transform import resize
from skimage.util import img_as_ubyte
import cv2
import numpy as np


def brightness(image, value=0):
    '''Adjust image brightness. Image is clipped to the range (0, 1)

    Args:
        image (np.array): image to have its brightness adjusted
        value (float): value to adjust all pixels by in the range (-1, 1).
                         (Defaults to 0)

    Returns:
        image (np.array): adjusted image
    '''
    if value == 0:
        return image
    return np.clip(image + value, 0, 1)


def binary_threshold(image, threshold=100):
    '''conver image to binary at a given threshold

    Args:
        image (np.array): image to be thresholded
        threshold (int): threshold value in the range (1, 255)
                     (Defaults to 100)

    Returns:
        image (np.array): thresholded image (float32 in range 0, 1)
    '''
    image = img_as_ubyte(image)
    if len(image.shape) > 2:
        if image.shape[2] == 3:
            image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    image = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, 11, int(threshold))
    if len(image.shape) == 2:
        image = np.stack((image,)*3, axis=-1)
    return np.clip(image.astype(np.float32) / 255.0, 0, 1)


def jpeg_compression(image, compression=101):
    '''encode image using jpeg then decode

    Args:
        image (np.array): image to be compressed
        compression (int): amount of jpeg compression, higher is better quality. 
                     101 returns identity image.
                     (Defaults to 101)

    Returns:
        image (np.array): jpeg compressed image
    '''
    if compression == 101:
        return image
    encode_param = [int(cv2.IMWRITE_JPEG_QUALITY), int(compression)]
    encoder = '.jpg'
    return _encode_compression(image, encoder, encode_param)


def _encode_compression(image, encoder, encode_param, uint=True):
    '''
        generic image encoder for jpeg, png etc
        using https://dev_resources/docs.opencv.org/3.4/d4/da8/group__imgcodecs.html to encode/decode
        for encode types: https://dev_resources/docs.opencv.org/3.4/d6/d87/imgcodecs_8hpp.html
    '''
    original_size = image.shape
    if uint == True:
        # jpeg/png encoder needs uint images not float
        image = img_as_ubyte(image)
    # encode image
    result, encoded_image = cv2.imencode(encoder, image, encode_param)
    # decode image
    decoded_image = cv2.imdecode(encoded_image, 1)
    if uint == True:
        # return to float type
        image = decoded_image.astype(np.float32) / 255.0
    # have to resize as jpeg can sometimes change the size of an image
    if image.shape != original_size:
        image = resize(image, original_size)
    return np.clip(image, 0, 1)


def contrast(image, contrast=1.0):
    """adjust the contrast of an image

    Args:
        image (np.array): image to have contrast adjusted
        contrast (float): amount of contrast to adjust by. 
            Values less that 1 will decrease contrast, 
            Values higher than 1 will increase the contrast.
            (Defaults to 1.0).

    Returns:
        image (np.array): image with contrast adjusted
    """
    if contrast == 1:
        return image
    image = img_as_ubyte(image)  # convert to 255 max value
    # need to zero center the data so we only adjust for contrast
    brightness = int(round(255*(1-contrast)/2))
    image = cv2.addWeighted(image, contrast, image, 0, brightness)
    return np.clip(image/255, 0, 1)


def hue(image, h=0):
    """adjust the hue of an image

    Args:
        image (np.array): image to have hue adjusted
        h (float): amount of hue to adjust by. 
            (Defaults to 0).

    Returns:
        image (np.array): image with hue adjusted
    """
    return _adjust_HSV(image, h, channel=0)


def saturation(image, sat=0):
    """adjust the saturation of an image

    Args:
        image (np.array): image to have saturation adjusted
        sat (float): amount of saturation to adjust by. 
            (Defaults to 0).

    Returns:
        image (np.array): image with saturation adjusted
    """
    return _adjust_HSV(image, sat, channel=1)


def brightness_hsv(image, b=0):
    """adjust the brightness of an image

    Args:
        image (np.array): image to have brightness adjusted
        b (float): amount of brightness to adjust by. 
            (Defaults to 0).

    Returns:
        image (np.array): image with brightness adjusted
    """
    return _adjust_HSV(image, b, channel=2)


def _adjust_HSV(image, value, channel):
    '''
    adjust hue, saturation or brightness of an image,
    image is float between 0, 1 and value is between -1 and 1
    channels:
        0 = hue
        1 = saturation
        2 = brightness
    '''
    if value == 0:
        return image
    image = img_as_ubyte(image)
    # convert to HSV
    hsv = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)
    # normalise between 0, 1
    hsv = hsv.astype(np.float32)
    hsv[:, :, 0] = hsv[:, :, 0]/179
    hsv[:, :, 1:] = hsv[:, :, 1:]/255
    # add the adjustment to the required channel
    hsv[:, :, channel] += value
    # account for hue colours being radial
    if channel == 0:
        hsv[:, :, channel] = np.mod(
            hsv[:, :, channel], 1)
    # reset to 0, 1 bounds
    hsv = np.clip(hsv, 0, 1)
    # convert back to RGB
    hsv[:, :, 0] = hsv[:, :, 0]*179
    hsv[:, :, 1:] = hsv[:, :, 1:]*255
    hsv = hsv.astype(np.uint8)
    return np.clip(cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)/255, 0, 1)


def blur(image, kernel_size=1):
    '''Gaussian Blur on an image

    Args:
        image (np.array): image to be rotated
        kernel_size (int): size of the convolutional kernel, will be converted
                           to an odd number. 1 is no blur.
                           (Defaults to 1)

    Returns:
        image (np.array): rotated image
    '''
    if kernel_size == 1:
        return image
    elif kernel_size > 0:
        blur_odd = (int(kernel_size/2)*2) + 1    # need to make kernel size odd
        image = cv2.GaussianBlur(
            image, (blur_odd, blur_odd), cv2.BORDER_DEFAULT)
        if len(image.shape) == 2:
            image = np.expand_dims(image, axis=2)
    return np.clip(image, 0, 1)



'''
UI create widgets
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import re
import os
from functools import partial
import datetime

import numpy as np
from PyQt6.QtWidgets import QPushButton, QLabel, QSlider, QCheckBox, QComboBox, QLineEdit
from PyQt6.QtGui import QIntValidator, QDoubleValidator, QValidator
from PyQt6.QtCore import Qt, pyqtSlot, QLocale

import IQM_Vis
from IQM_Vis.UI.custom_widgets import ClickLabel
from IQM_Vis.utils import gui_utils, plot_utils, image_utils, save_utils

# sub class used by IQM_Vis.main.make_app to initialise widgets and general UI functions for widgets
class widgets():
    def init_widgets(self, **kwargs):
        '''
        create all the widgets we need and init params
        '''
        self._init_widgets()
        self.set_image_name_text()
        self._init_image_settings()

    def _init_widgets(self):
        '''
        create all the widgets we need and init params
        '''
        self.export_in_progress = False
        self.last_export = None
        self.tool_tip_style = """QToolTip {background-color: black; color: white; border: black solid 1px}"""
        # first setup the slider data
        self.sliders = {'transforms': {}, 'metric_params': {}}
        self._init_sliders(self.sliders['transforms'], self.checked_transformations,  param_group='transforms')
        self._init_sliders(self.sliders['metric_params'], self.metric_params, param_group='metric_params')

        self.widget_row = {}
        for i, data_store in enumerate(self.data_stores):   # create each row of the widgets
            self.widget_row[i] = {'images':{}, 'metrics':{}, 'metric_images':{}}
            '''image and transformed image'''
            for image_name in ['original', 'transformed']:
                self.widget_row[i]['images'][image_name] = {}
                # image widget
                self.widget_row[i]['images'][image_name]['data'] = QLabel(self)
                self.widget_row[i]['images'][image_name]['data'].setAlignment(Qt.AlignmentFlag.AlignCenter)
                # image label
                self.widget_row[i]['images'][image_name]['label'] = QLabel(self)
                self.widget_row[i]['images'][image_name]['label'].setAlignment(Qt.AlignmentFlag.AlignCenter)
                self.widget_row[i]['images'][image_name]['label'].setText(image_name)

            '''metric images'''
            for metric_image in data_store.metric_images:
                if metric_image in self.checked_metric_images:
                    self.widget_row[i]['metric_images'][metric_image] = {}
                    self.widget_row[i]['metric_images'][metric_image]['data'] = QLabel(self)
                    self.widget_row[i]['metric_images'][metric_image]['data'].setAlignment(Qt.AlignmentFlag.AlignCenter)
                    # image label
                    self.widget_row[i]['metric_images'][metric_image]['label'] = QLabel(self)
                    self.widget_row[i]['metric_images'][metric_image]['label'].setAlignment(Qt.AlignmentFlag.AlignCenter)

            '''metrics graphs'''
            im_pair_name = gui_utils.get_image_pair_name(data_store)
            self.widget_row[i]['metrics']['info'] = {}
            self.widget_row[i]['metrics']['info']['label'] = QLabel(self)
            self.widget_row[i]['metrics']['info']['label'].setAlignment(Qt.AlignmentFlag.AlignCenter)
            self.widget_row[i]['metrics']['info']['label'].setText('Metrics '+im_pair_name)
            if self.metrics_info_format == 'graph':
                self.widget_row[i]['metrics']['info']['data'] = gui_utils.MplCanvas(size=(self.graph_size/10, self.graph_size/10))
            else:
                self.widget_row[i]['metrics']['info']['data'] = QLabel(self)
                self.widget_row[i]['metrics']['info']['data'].setAlignment(Qt.AlignmentFlag.AlignCenter)
                self.widget_row[i]['metrics']['info']['data'].setText('')
            # metrics average graphs
            if self.metrics_avg_graph:
                self.widget_row[i]['metrics']['avg'] = {}
                self.widget_row[i]['metrics']['avg']['label'] = QLabel(self)
                self.widget_row[i]['metrics']['avg']['label'].setAlignment(Qt.AlignmentFlag.AlignCenter)
                self.widget_row[i]['metrics']['avg']['label'].setText('IQM Averages')
                self.widget_row[i]['metrics']['avg']['data'] = gui_utils.MplCanvas(size=(self.graph_size/10, self.graph_size/10), polar=True)
                self.widget_row[i]['metrics']['avg']['data'].setToolTip('Mean metric value over the range of each transform.')
                self.widget_row[i]['metrics']['avg']['data'].setStyleSheet(self.tool_tip_style)
                self.widget_row[i]['metrics']['avg']['save_button'] = QPushButton('Adjust or Save', self)
                self.widget_row[i]['metrics']['avg']['save_button'].clicked.connect(partial(self.plot_radar_mlp, i))
            if self.metric_range_graph:
                self.widget_row[i]['metrics']['range'] = {}
                self.widget_row[i]['metrics']['range']['label'] = QLabel(self)
                self.widget_row[i]['metrics']['range']['label'].setAlignment(Qt.AlignmentFlag.AlignCenter)
                self.widget_row[i]['metrics']['range']['label'].setText('Response Profiles')
                self.widget_row[i]['metrics']['range']['data'] = gui_utils.MplCanvas(size=(self.graph_size/10, self.graph_size/10))
                self.widget_row[i]['metrics']['range']['data'].setToolTip('Single transformation value range for all metrics.')
                self.widget_row[i]['metrics']['range']['data'].setStyleSheet(self.tool_tip_style)
                self.widget_row[i]['metrics']['range']['save_button'] = QPushButton('Adjust or Save', self)
                self.widget_row[i]['metrics']['range']['save_button'].clicked.connect(partial(self.plot_metric_range_mlp, i))
            # correlation graphs
            self.widget_row[i]['metrics']['correlation'] = {}
            self.widget_row[i]['metrics']['correlation']['label'] = QLabel(self)
            self.widget_row[i]['metrics']['correlation']['label'].setAlignment(Qt.AlignmentFlag.AlignCenter)
            self.widget_row[i]['metrics']['correlation']['label'].setText('Human Correlation')
            self.widget_row[i]['metrics']['correlation']['data'] = gui_utils.MplCanvas(size=(self.graph_size/10, self.graph_size/10))
            self.widget_row[i]['metrics']['correlation']['data'].setToolTip('Human scores versus IQMs.\nMean shown with points\nStandard deviation shown with bars.\nClick points to show image.')
            self.widget_row[i]['metrics']['correlation']['data'].setStyleSheet(self.tool_tip_style)
            # JND graphs
            self.widget_row[i]['metrics']['JND'] = {}
            self.widget_row[i]['metrics']['JND']['label'] = QLabel(self)
            self.widget_row[i]['metrics']['JND']['label'].setAlignment(Qt.AlignmentFlag.AlignCenter)
            self.widget_row[i]['metrics']['JND']['label'].setText('Just Noticeable Difference')
            self.widget_row[i]['metrics']['JND']['data'] = gui_utils.MplCanvas(size=(self.graph_size/10, self.graph_size/10))
            self.widget_row[i]['metrics']['JND']['data'].setToolTip(
                'Just Noticeable Difference shows the perceivable effect of a distortion for a human subject.')
            self.widget_row[i]['metrics']['JND']['data'].setStyleSheet(self.tool_tip_style)


        '''buttons'''
        self.widget_controls = {'button': {}, 'slider': {}, 'label': {}, 'check_box': {}, 'images': {}}
        
        if (self.metrics_avg_graph or self.metric_range_graph):
            # Update graphs
            self.widget_controls['button']['force_update'] = QPushButton('Update Graphs', self)
            self.widget_controls['button']['force_update'].setToolTip('Update graphs using all the current slider values.')
            self.widget_controls['button']['force_update'].setStyleSheet(self.tool_tip_style)
            self.widget_controls['button']['force_update'].clicked.connect(self.display_images)
            self.widget_controls['button']['force_update'].clicked.connect(partial(self.redo_plots, True))
            # Change plot limits
            self.widget_controls['check_box']['graph_limits'] = QCheckBox('Squeeze plots')
            self.widget_controls['check_box']['graph_limits'].setToolTip('Set the scale of the plots to the metric data range.')
            self.widget_controls['check_box']['graph_limits'].setStyleSheet(self.tool_tip_style)
            self.widget_controls['check_box']['graph_limits'].setCheckState(Qt.CheckState.Unchecked)
            self.widget_controls['check_box']['graph_limits'].stateChanged.connect(self.change_plot_lims)

        if self.metric_range_graph:
            # buttons to control which metric range graph to show
            self.widget_controls['button']['next_metric_graph'] = QPushButton('->', self)
            self.widget_controls['button']['next_metric_graph'].clicked.connect(partial(self.change_metric_range_graph, 1))
            self.widget_controls['button']['prev_metric_graph'] = QPushButton('<-', self)
            self.widget_controls['button']['prev_metric_graph'].clicked.connect(partial(self.change_metric_range_graph, -1))
        # buttons to control which correlation graph to show
        self.widget_controls['button']['next_correlation_graph'] = QPushButton('->', self)
        self.widget_controls['button']['next_correlation_graph'].clicked.connect(partial(self.change_metric_correlations_graph, 1))
        self.widget_controls['button']['prev_correlation_graph'] = QPushButton('<-', self)
        self.widget_controls['button']['prev_correlation_graph'].clicked.connect(partial(self.change_metric_correlations_graph, -1))

        if self.dataset:
            # control what image is used from the dataset
            self.widget_controls['button']['next_data'] = QPushButton('->', self)
            self.widget_controls['button']['next_data'].clicked.connect(partial(self.change_preview_images, 1))
            self.widget_controls['button']['prev_data'] = QPushButton('<-', self)
            self.widget_controls['button']['prev_data'].clicked.connect(partial(self.change_preview_images, -1))
            self.widget_controls['label']['data'] = QLabel(self)
            self.widget_controls['label']['data'].setText('Change Image:')
            self.widget_controls['label']['data_num'] = QLabel(self)
            self.widget_controls['label']['data_num'].setText('1/1')
            self.widget_im_num_hash = {}
            for im in range(self.num_images_scroll_show):
                self.widget_controls['images'][im] = QLabel( self)
                self.widget_controls['images'][im].setAlignment(Qt.AlignmentFlag.AlignCenter)
                self.widget_controls['images'][im].mousePressEvent = partial(self.change_data_click_im, im)
                self.widget_im_num_hash[im] = im # store the ind number of each im widget preview

        # launch experiment buttons
        self.widget_controls['button']['launch_exp_2AFC'] = QPushButton('Run 2AFC Experiment', self)
        self.widget_controls['button']['launch_exp_2AFC'].setToolTip('Two Alternative Forced Choice Experiment')
        self.widget_controls['button']['launch_exp_2AFC'].clicked.connect(self.launch_experiment_2AFC)
        self.widget_controls['button']['launch_exp_JND'] = QPushButton('Run JND Experiment', self)
        self.widget_controls['button']['launch_exp_JND'].setToolTip('Just Noticeable Difference Experiment')
        self.widget_controls['button']['launch_exp_JND'].clicked.connect(self.launch_experiment_JND)
        # load 2AFC experiment button
        self.widget_controls['button']['load_exp'] = QPushButton('Load Experiment', self)
        self.widget_controls['button']['load_exp'].clicked.connect(self.load_experiment_from_dir)
        # load JNDexperiment button
        self.widget_controls['button']['load_exp_JND'] = QPushButton('Load Experiment', self)
        self.widget_controls['button']['load_exp_JND'].clicked.connect(self.load_experiment_from_dir_JND)
        # export images button
        self.widget_controls['button']['export_images'] = QPushButton('Export Images', self)
        self.widget_controls['button']['export_images'].clicked.connect(self.export_trans_images)
        


        '''sliders'''
        self.params_from_sliders = {}
        for param_group, slider_group in self.sliders.items():
            self.params_from_sliders[param_group] = {}
            self.widget_controls['button'][param_group] = {}
            self.widget_controls['button'][param_group]['reset_sliders'] = QPushButton('Reset', self)
            if param_group == 'metric_params':
                redo_graphs = True
            else:
                redo_graphs = False
            self.widget_controls['button'][param_group]['reset_sliders'].clicked.connect(
                partial(self.reset_slider_group, param_group, redo_graphs))

            for key, item_sliders in slider_group.items():
                self.widget_controls['slider'][key] = {}
                self.widget_controls['slider'][key]['data'] = QSlider(Qt.Orientation.Horizontal)
                self.widget_controls['slider'][key]['data'].setMinimum(0)
                self.widget_controls['slider'][key]['data'].setMaximum(len(item_sliders['values'])-1)
                for func in item_sliders['value_change']:
                    self.widget_controls['slider'][key]['data'].valueChanged.connect(func)
                for func in item_sliders['release']:
                    self.widget_controls['slider'][key]['data'].sliderReleased.connect(func)
                self.params_from_sliders['transforms'][key] = item_sliders['init_ind']
                self.widget_controls['slider'][key]['label'] = QLabel(self)
                self.widget_controls['slider'][key]['label'].setAlignment(Qt.AlignmentFlag.AlignRight)
                self.widget_controls['slider'][key]['label'].setText(f"{key}:")
                self.widget_controls['slider'][key]['value'] = QLabel(self)
                self.widget_controls['slider'][key]['value'].setAlignment(Qt.AlignmentFlag.AlignLeft)
                self.widget_controls['slider'][key]['value'].setText(str(self.params_from_sliders['transforms'][key]))

        ''' experiment options '''
        self.widget_experiment_params = {}
        
        for trans_name, deets in self.checked_transformations.items():
            self.widget_experiment_params[trans_name] = {}
            self.widget_experiment_params[trans_name]['check_box'] = QCheckBox(self)
            self.widget_experiment_params[trans_name]['check_box'].setChecked(True)
            self.widget_experiment_params[trans_name]['check_box'].stateChanged.connect(partial(self.change_text_exp_trans, 
                                                                                                  trans=trans_name))

            self.widget_experiment_params[trans_name]['name'] = QLabel(self)
            self.widget_experiment_params[trans_name]['name'].setText(trans_name)

            self.widget_experiment_params[trans_name]['min'] = QLabel(self)
            self.widget_experiment_params[trans_name]['min'].setText('min:')
            self.widget_experiment_params[trans_name]['min_edit'] = QLineEdit()
            self.widget_experiment_params[trans_name]['min_edit'].setValidator(get_float_validator())
            self.widget_experiment_params[trans_name]['min_edit'].textChanged.connect(partial(self.colour_lineedit, self.widget_experiment_params[trans_name]['min_edit']))
            self.widget_experiment_params[trans_name]['min_edit'].setText(f"{deets['min']}")

            self.widget_experiment_params[trans_name]['max'] = QLabel(self)
            self.widget_experiment_params[trans_name]['max'].setText('max:')
            self.widget_experiment_params[trans_name]['max_edit'] = QLineEdit()
            self.widget_experiment_params[trans_name]['max_edit'].setValidator(get_float_validator())
            self.widget_experiment_params[trans_name]['max_edit'].textChanged.connect(partial(self.colour_lineedit, self.widget_experiment_params[trans_name]['max_edit']))
            self.widget_experiment_params[trans_name]['max_edit'].setText(f"{deets['max']}")

            self.widget_experiment_params[trans_name]['steps'] = QLabel(self)
            self.widget_experiment_params[trans_name]['steps'].setText('steps:')
            self.widget_experiment_params[trans_name]['steps_edit'] = QLineEdit()
            self.widget_experiment_params[trans_name]['steps_edit'].setValidator(QIntValidator())
            self.widget_experiment_params[trans_name]['steps_edit'].setMaxLength(2)
            self.widget_experiment_params[trans_name]['steps_edit'].setText(f"{self.num_step_experiment}")
        # change save folder button
        self.widget_controls['label']['exp_change_save'] = QLabel(self)
        self.widget_controls['label']['exp_change_save'].setText(f'Root Save Folder: {self.default_save_dir}')
        self.widget_controls['button']['exp_change_save'] = QPushButton('Change', self)
        self.widget_controls['button']['exp_change_save'].clicked.connect(self.change_save_folder)
        # dataset name
        self.widget_controls['label']['exp_change_dataset_name'] = QLabel(self)
        self.widget_controls['label']['exp_change_dataset_name'].setText(f'Dataset Name (JND)')
        self.widget_controls['button']['exp_change_dataset_name'] = QLineEdit(self)
        self.widget_controls['button']['exp_change_dataset_name'].setText(
            self.default_dataset_name)
        self.widget_controls['button']['exp_change_dataset_name'].textChanged.connect(
            self.change_dataset_name)
        self.widget_controls['label']['exp_im_nums'] = QLabel(self)
        self.widget_controls['label']['exp_im_nums'].setText('Dataset Range (JND):')
        self.widget_controls['label']['exp_im_nums_dash'] = QLabel(self)
        self.widget_controls['label']['exp_im_nums_dash'].setText('-')
        self.widget_controls['label']['range_edit_lower'] = QLineEdit()
        self.widget_controls['label']['range_edit_lower'].setValidator(QIntValidator())
        self.exp_range_lower = 1
        self.widget_controls['label']['range_edit_lower'].setText(
            f"{self.exp_range_lower}")
        self.widget_controls['label']['range_edit_lower'].textChanged.connect(self.JND_dataset_range_lower)
        self.widget_controls['label']['range_edit_upper'] = QLineEdit()
        self.widget_controls['label']['range_edit_upper'].setValidator(QIntValidator())
        self.exp_range_upper = len(self.data_stores[0])
        self.widget_controls['label']['range_edit_upper'].setText(
            f"{self.exp_range_upper}")
        self.widget_controls['label']['range_edit_upper'].textChanged.connect(self.JND_dataset_range_upper)

        ''' export options '''
        self.widget_export = {}
        
        for trans_name, deets in self.checked_transformations.items():
            self.widget_export[trans_name] = {}
            self.widget_export[trans_name]['check_box'] = QCheckBox(self)
            self.widget_export[trans_name]['check_box'].setChecked(True)
            self.widget_export[trans_name]['check_box'].stateChanged.connect(partial(self.change_text_export_trans,
                                                                                                  trans=trans_name))

            self.widget_export[trans_name]['name'] = QLabel(self)
            self.widget_export[trans_name]['name'].setText(trans_name)

            self.widget_export[trans_name]['min'] = QLabel(self)
            self.widget_export[trans_name]['min'].setText('min:')
            self.widget_export[trans_name]['min_edit'] = QLineEdit()
            self.widget_export[trans_name]['min_edit'].setValidator(get_float_validator())
            self.widget_export[trans_name]['min_edit'].textChanged.connect(partial(self.colour_lineedit, self.widget_export[trans_name]['min_edit']))
            self.widget_export[trans_name]['min_edit'].setText(f"{deets['min']}")

            self.widget_export[trans_name]['max'] = QLabel(self)
            self.widget_export[trans_name]['max'].setText('max:')
            self.widget_export[trans_name]['max_edit'] = QLineEdit()
            self.widget_export[trans_name]['max_edit'].setValidator(get_float_validator())
            self.widget_export[trans_name]['max_edit'].textChanged.connect(partial(self.colour_lineedit, self.widget_export[trans_name]['max_edit']))
            self.widget_export[trans_name]['max_edit'].setText(f"{deets['max']}")

            self.widget_export[trans_name]['steps'] = QLabel(self)
            self.widget_export[trans_name]['steps'].setText('steps:')
            self.widget_export[trans_name]['steps_edit'] = QLineEdit()
            self.widget_export[trans_name]['steps_edit'].setValidator(QIntValidator())
            self.widget_export[trans_name]['steps_edit'].setMaxLength(2)
            self.widget_export[trans_name]['steps_edit'].setText(f"{self.num_steps_range}")
        # change save folder button
        self.widget_controls['label']['export_change_save'] = QLabel(self)
        self.widget_controls['label']['export_change_save'].setText(f'Save Folder: {self.default_save_dir}')
        self.widget_controls['button']['export_change_save'] = QPushButton('Change', self)
        self.widget_controls['button']['export_change_save'].clicked.connect(self.change_save_folder)


    '''
    setup/helper functions
    '''
    def _init_sliders(self, sliders_dict, info_dict, param_group):
        '''
        generic initialiser of sliders (will change the dicts input rather than returning - like C++ pointers &)
            sliders_dict: holds the slider widgets
            info_dict: info to initialise the sliders
        '''
        for key, info_item in info_dict.items():
            sliders_dict[key] = {}
            sliders_dict[key]['release'] = [self.display_images]
            sliders_dict[key]['value_change'] = [partial(self.generic_value_change, key, param_group), self.display_images]
            if 'function' in info_item.keys():
                sliders_dict[key]['function'] = info_item['function']
            if 'init_value' not in info_item.keys():
                info_item['init_value'] = 0
            sliders_dict[key]['init_value_store'] = info_item['init_value']
            if 'values' in info_item.keys():
                sliders_dict[key]['values'] = info_item['values']
            else:
                if 'num_values' not in info_item.keys():
                    info_item['num_values'] = 41   # make default value for steps in slider range
                # see if we need to make odd numbers (for use with kernel sizes)
                if 'normalise' in info_item.keys():
                    # store for if min/max changed later
                    sliders_dict[key]['normalise'] = info_item['normalise']
                self.make_slider_range(sliders_dict, key, info_item['min'], info_item['max'], info_item['num_values'])
            
            # store num steps used
            sliders_dict[key]['default_num_steps'] = len(sliders_dict[key]['values'])
            # make min/max inputs
            width = 40
            sliders_dict[key]['min_edit'] = QLineEdit()
            sliders_dict[key]['min_edit'].setValidator(get_float_validator())
            sliders_dict[key]['min_edit'].textChanged.connect(partial(self.colour_lineedit, sliders_dict[key]['min_edit']))
            sliders_dict[key]['min_edit'].setText(f"{min(sliders_dict[key]['values'])}")
            sliders_dict[key]['min_edit'].setFixedWidth(width)
            sliders_dict[key]['min_edit'].editingFinished.connect(partial(self.edit_slider_vals, sliders_dict, key, info_item))
            sliders_dict[key]['min_edit'].editingFinished.connect(partial(self.generic_value_change, key, param_group))
            sliders_dict[key]['min_edit'].editingFinished.connect(self.display_images)
            sliders_dict[key]['max_edit'] = QLineEdit()
            sliders_dict[key]['max_edit'].setValidator(get_float_validator())
            sliders_dict[key]['max_edit'].textChanged.connect(partial(self.colour_lineedit, sliders_dict[key]['max_edit']))
            sliders_dict[key]['max_edit'].setText(f"{max(sliders_dict[key]['values'])}")
            sliders_dict[key]['max_edit'].setFixedWidth(width)
            sliders_dict[key]['max_edit'].editingFinished.connect(partial(self.edit_slider_vals, sliders_dict, key, info_item))
            sliders_dict[key]['max_edit'].editingFinished.connect(partial(self.generic_value_change, key, param_group))
            sliders_dict[key]['max_edit'].editingFinished.connect(self.display_images)
    
    def make_slider_range(self, sliders_dict, key, _min, _max, num_steps=None):
        if num_steps == None:
            num_steps = sliders_dict[key]['default_num_steps']
        sliders_dict[key]['values'] = np.linspace(_min, _max, num_steps)
        # see if we need to make odd numbers (for use with kernel sizes)
        if 'normalise' in sliders_dict[key].keys():
            if sliders_dict[key]['normalise'] == 'odd':
                sliders_dict[key]['values'] = np.round(sliders_dict[key]['values'])
                sliders_dict[key]['values'] = sliders_dict[key]['values'][sliders_dict[key]['values']%2 == 1]
        # get ind of the initial value to set the slider at
        sliders_dict[key]['init_ind'] = np.searchsorted(sliders_dict[key]['values'], sliders_dict[key]['init_value_store'], side='left')

    def _init_image_settings(self):
        self.widget_settings = {}

        # pre processing
        self.pre_processing_options = {'None': None,
                                       'Resize 32': partial(IQM_Vis.utils.image_utils.resize_to_longest_side, side=32),
                                       'Resize 64': partial(IQM_Vis.utils.image_utils.resize_to_longest_side, side=64),
                                       'Resize 128': partial(IQM_Vis.utils.image_utils.resize_to_longest_side, side=128),
                                       'Resize 256': partial(IQM_Vis.utils.image_utils.resize_to_longest_side, side=256),
                                       'Resize 512': partial(IQM_Vis.utils.image_utils.resize_to_longest_side, side=512),
                                       'Resize 1280': partial(IQM_Vis.utils.image_utils.resize_to_longest_side, side=1280),
                                       'Resize 1920': partial(IQM_Vis.utils.image_utils.resize_to_longest_side, side=1920),
                                       'Resize 2560': partial(IQM_Vis.utils.image_utils.resize_to_longest_side, side=2560),
                                       'Resize 3840': partial(IQM_Vis.utils.image_utils.resize_to_longest_side, side=3840),
                                       }
        if not hasattr(self, 'pre_processing_option'):
            self.pre_processing_option = 'Resize 256'  # init_val

        for i, data_store in enumerate(self.data_stores):
            if hasattr(data_store, 'image_pre_processing'):
                if str(data_store.image_pre_processing) not in [str(f) for f in self.pre_processing_options.values()]:
                    name = f"Custom {i}"
                    self.pre_processing_options[name] = data_store.image_pre_processing
                    self.pre_processing_option = name
        combobox_pre = QComboBox()
        combobox_pre.addItems(list(self.pre_processing_options.keys()))
        combobox_pre.setCurrentText(self.pre_processing_option)
        combobox_pre.activated.connect(self.enable_settings_button)
        self.widget_settings['image_pre_processing'] = {'widget': combobox_pre, 'label': QLabel('Image Pre Processing:')}
        self.change_pre_processing()   # apply default

        # post processing
        self.post_processing_options = {'None': None,
                                        'Crop Centre': partial(IQM_Vis.utils.image_utils.crop_centre, scale_factor=2, keep_size=True),
                                        'Crop Centre (No Resize)': partial(IQM_Vis.utils.image_utils.crop_centre, scale_factor=2, keep_size=False),
                                        }
        if not hasattr(self, 'post_processing_option'):
            self.post_processing_option = 'None'  # init_val
        for i, data_store in enumerate(self.data_stores):
            if hasattr(data_store, 'image_post_processing'):
                if data_store.image_post_processing not in list(self.post_processing_options.values()):
                    name = f"Custom {i}"
                    self.post_processing_options[name] = data_store.image_post_processing
                    self.post_processing_option = name
        combobox_post = QComboBox()
        combobox_post.addItems(list(self.post_processing_options.keys()))
        combobox_post.setCurrentText(self.post_processing_option)
        combobox_post.activated.connect(self.enable_settings_button)
        self.widget_settings['image_post_processing'] = {'widget': combobox_post, 'label': QLabel('Image Post Processing:')}
        self.change_post_processing()    # apply default

        # image display size
        line_edit_display = QLineEdit()
        line_edit_display.setValidator(QIntValidator())
        line_edit_display.setMaxLength(3)
        line_edit_display.setText(str(self.image_display_size))
        line_edit_display.textChanged.connect(self.enable_settings_button)
        self.widget_settings['image_display_size'] = {'widget': line_edit_display, 'label': QLabel('Image Display Size:')}

        # graph display size
        line_edit_graph = QLineEdit()
        line_edit_graph.setValidator(QIntValidator())
        line_edit_graph.setMaxLength(2)
        line_edit_graph.setText(str(self.graph_size))
        line_edit_graph.textChanged.connect(self.enable_settings_button)
        self.widget_settings['graph_display_size'] = {'widget': line_edit_graph, 'label': QLabel('Graph Display Size:')}

        # graph/experiment number of steps in the range
        line_edit_num_steps = QLineEdit()
        line_edit_num_steps.setValidator(QIntValidator())
        line_edit_num_steps.setMaxLength(4)
        line_edit_num_steps.setText(str(self.num_steps_range))
        line_edit_num_steps.textChanged.connect(self.enable_settings_button)
        self.widget_settings['graph_num_steps'] = {'widget': line_edit_num_steps, 'label': QLabel('Graph Range Step Size:')}

        # image screen calibration
        line_edit_rgb = QLineEdit()
        line_edit_rgb.setValidator(QIntValidator())
        line_edit_rgb.setMaxLength(4)
        line_edit_rgb.setText(str(self.rgb_brightness))
        line_edit_rgb.textChanged.connect(self.enable_settings_button)
        self.widget_settings['image_display_rgb_brightness'] = {
            'widget': line_edit_rgb, 'label': QLabel('RGB Max Brightness:')}

        line_edit_display = QLineEdit()
        line_edit_display.setValidator(QIntValidator())
        line_edit_display.setMaxLength(4)
        line_edit_display.setText(str(self.display_brightness))
        line_edit_display.textChanged.connect(self.enable_settings_button)
        self.widget_settings['image_display_display_brightness'] = {
            'widget': line_edit_display, 'label': QLabel('Display Max Brightness:')}

        # update settings button
        self.widget_settings['update_button'] = QPushButton('Apply Settings', self)
        # self.settings_text_colour_original = self.widget_settings['update_button'].palette().color(
        #     QPalette.ColorRole.Text).name()   # this doesn't work for some reason ...
        self.settings_text_colour_original = 'black'
        self.widget_settings['update_button'].clicked.connect(self.update_image_settings)
        self.disable_settings_button()

    '''
    ==================== functions to bind to sliders/widgets ====================
    '''
    # line edit colour based on if valid
    def colour_lineedit(self, widget, txt):
        if not is_float(txt):
            widget.setStyleSheet("color: red;")
        else:
            widget.setStyleSheet("color: black;")

    # sliders value changes
    def generic_value_change(self, key, param_group):
        index = self.widget_controls['slider'][key]['data'].value()
        self.params_from_sliders[param_group][key] = self.sliders[param_group][key]['values'][index]
        self.display_slider_num(key, param_group) # display the new value ont UI

    def edit_slider_vals(self, sliders_dict, key, info_item):
        _min = make_float_from_text(sliders_dict[key]['min_edit'].text())
        _max = make_float_from_text(sliders_dict[key]['max_edit'].text())
        self.make_slider_range(sliders_dict, key, _min, _max)
        info_item['min'] = _min
        info_item['max'] = _max

    def display_slider_num(self, key, param_group, disp_len=5):
        # display the updated value
        value_str = str(self.params_from_sliders[param_group][key])
        value_str = gui_utils.str_to_len(value_str, disp_len, '0', plus=True)
        self.widget_controls['slider'][key]['value'].setText(value_str)

    def reset_slider_group(self, param_group, redo_plots=False, display_images=True):
        for key, item_sliders in self.sliders[param_group].items():
            self.widget_controls['slider'][key]['data'].setValue(item_sliders['init_ind'])
        if display_images == True:
            self.display_images()
        if redo_plots == True:
            self.redo_plots(calc_range=False)

    def reset_sliders(self):
        self.update_images = False   # dont calc any images/metrics when updating sliders now is slow
        for param_group in self.sliders:
            self.reset_slider_group(param_group, False, False)
        self.update_images = True  # make sure to return to calc images/metrics
        self.display_images()
        self.redo_plots(calc_range=False)

    def set_image_name_text(self):
        for i, data_store in enumerate(self.data_stores):
            res = gui_utils.get_resolutions(data_store)
            self.widget_row[i]['images']['original']['label'].setText(f"{data_store.get_reference_image_name()}\n{res['reference']}")
            self.widget_row[i]['images']['transformed']['label'].setText(f"{gui_utils.get_transformed_image_name(data_store)}\n{res['transform']}")
            for metric_image in data_store.metric_images:
                if metric_image in self.checked_metric_images:
                    metric_name = gui_utils.get_metric_image_name(metric_image, data_store)
                    if len(metric_name) > 20:
                        metric_name = metric_image
                    self.widget_row[i]['metric_images'][metric_image]['label'].setText(metric_name)

    def change_plot_lims(self, state):
        if state == 2:  # 2 is the checked value
            self.plot_data_lim = self.data_lims['range_data']
        else:
            self.plot_data_lim = self.data_lims['fixed']
        self.redo_plots(calc_range=False)


    def change_dataset_name(self, txt):
            ''' change the dataset_name we are using '''
            self.default_dataset_name = txt

    def JND_dataset_range_lower(self, txt):
        ''' change the dataset_name we are using '''
        try:
            self.exp_range_lower = int(txt)
        except:
            return  
        if self.exp_range_lower < 1:
            self.exp_range_lower = 1
        elif self.exp_range_lower > self.exp_range_upper:
            self.exp_range_lower = self.exp_range_upper
        self.widget_controls['label']['range_edit_lower'].setText(
            f"{self.exp_range_lower}")

    def JND_dataset_range_upper(self, txt):
        ''' change the dataset_name we are using '''
        try:
            self.exp_range_upper = int(txt)
        except:
            return
        if self.exp_range_upper < self.exp_range_lower:
            self.exp_range_upper = self.exp_range_lower
        elif self.exp_range_upper > len(self.data_stores[0]):
            self.exp_range_upper = len(self.data_stores[0])
        self.widget_controls['label']['range_edit_upper'].setText(
            f"{self.exp_range_upper}")
    '''
    images settings apply
    '''
    def enable_settings_button(self):
        self.widget_settings['update_button'].setEnabled(True)
        if hasattr(self, 'settings_text_colour_original'):
            self.widget_settings['update_button'].setStyleSheet(
                f"QPushButton {{color: {self.settings_text_colour_original};}}")

    def disable_settings_button(self):
        self.widget_settings['update_button'].setEnabled(False)
        self.widget_settings['update_button'].setStyleSheet(f"QPushButton {{color: gray;}}")

    
    def update_image_settings(self):
        ''' button to apply new image settings '''
        if hasattr(self, 'range_worker'):
            self.range_worker.stop()
        # apply changes to settings
        self.change_pre_processing()
        self.change_post_processing()
        self.change_display_im_size()
        self.change_graph_size()
        self.change_num_steps()
        self.change_display_im_rgb_brightness()
        self.change_display_im_display_brightness()

        self.change_data(0, _redo_plots=False)  # update the image data settings
        self.display_images()
        if hasattr(self, 'update_UI'):
            if self.update_UI == True:
                self.construct_UI()
        if hasattr(self, 'image_settings_update_plots'):
            if self.image_settings_update_plots == True:
                # only redo the graphs if nessesary
                self.redo_plots(calc_range=True)
        self.image_settings_update_plots = False
        self.update_UI = False
        self.disable_settings_button()

    '''
    image settings setters
    '''

    def change_pre_processing(self, *args):
        self.pre_processing_option = self.widget_settings['image_pre_processing']['widget'].currentText(
        )
        for data_store in self.data_stores:
            if hasattr(data_store, 'image_pre_processing'):
                if data_store.image_pre_processing != self.pre_processing_options[self.pre_processing_option]:
                    data_store.image_pre_processing = self.pre_processing_options[
                        self.pre_processing_option]
                    self.image_settings_update_plots = True

    def change_post_processing(self, *args):
        self.post_processing_option = self.widget_settings['image_post_processing']['widget'].currentText(
        )
        for data_store in self.data_stores:
            if hasattr(data_store, 'image_post_processing'):
                if data_store.image_post_processing != self.post_processing_options[self.post_processing_option]:
                    data_store.image_post_processing = self.post_processing_options[
                        self.post_processing_option]
                    self.image_settings_update_plots = True
        # self.display_images()
        # self.redo_plots(calc_range=False)

    def change_display_im_size(self):
        txt = self.widget_settings['image_display_size']['widget'].text()
        if txt == '':
            txt = 1
        new_size = max(1, int(txt))
        if new_size != self.image_display_size:
            self.image_display_size = new_size
            self.update_UI = True
        # self.construct_UI()
        # self.display_images()
        # if old_size > self.image_display_size:
        #     self.setMaximumSize(self.main_widget.sizeHint())
        # if old_size < self.image_display_size:
        #     self.setMinimumSize(self.main_widget.sizeHint())

    def change_graph_size(self):
        txt = self.widget_settings['graph_display_size']['widget'].text()
        if txt == '':
            txt = 1
        new_size = max(1, int(txt))
        if new_size != self.graph_size:
            self.graph_size = new_size
            self.update_UI = True

    def change_num_steps(self):
        txt = self.widget_settings['graph_num_steps']['widget'].text()
        if txt == '':
            txt = 1
        new_steps = max(2, int(txt))
        if new_steps != self.num_steps_range:
            self.num_steps_range = new_steps
            self.update_UI = True
    
    def change_display_im_rgb_brightness(self):
        txt = self.widget_settings['image_display_rgb_brightness']['widget'].text()
        if txt == '':
            txt = 1
        self.rgb_brightness = max(1, int(txt))
    
    def change_display_im_display_brightness(self):
        txt = self.widget_settings['image_display_display_brightness']['widget'].text()
        if txt == '':
            txt = 1
        self.display_brightness = max(1, int(txt))

    '''
    status bar
    '''
    def update_progress(self, v):
        self.pbar.setValue(v)
        if v == 0:
            self.status_bar.showMessage('Done', 3000)

    def update_status_bar(self, v, time=None):
        if time == None:
            self.status_bar.showMessage(v)
        elif isinstance(time, int):
            self.status_bar.showMessage(v, time)

    '''
    export controls
    '''
    def change_text_export_trans(self, trans):
        ''' change colour of text when checkbox is pressed '''
        checked = self.widget_export[trans]['check_box'].isChecked()
        for widget in self.widget_export[trans]:
            if widget == 'check_box':
                pass
            else:
                self.widget_export[trans][widget].setEnabled(checked)
                if checked == True:
                    self.widget_export[trans][widget].setStyleSheet(f"QLineEdit {{color: {self.settings_text_colour_original};}}\nQLabel {{color: {self.settings_text_colour_original};}}")
                else:
                    self.widget_export[trans][widget].setStyleSheet(f"QLineEdit {{color: gray;}}\nQLabel {{color: gray;}}")

    def export_trans_images(self):
        self.export_in_progress = True
        # make save folder
        dir_folder = self.get_export_dir(0)
        save_folder = os.path.join(
            dir_folder, f"images-{str(datetime.datetime.now()).split('.')[0]}")
        if not os.path.exists(save_folder):
            os.makedirs(save_folder)

        # save original image
        ref_image = self.data_stores[0].get_reference_image()
        image_utils.save_image(ref_image,
                               os.path.join(save_folder, f"reference.png"))

        # make and save transforms
        checked_transforms = {}
        for trans in self.widget_export:
            if self.widget_export[trans]['check_box'].isChecked():
                data = {'min': make_float_from_text(self.widget_export[trans]['min_edit'].text()),
                        'max': make_float_from_text(self.widget_export[trans]['max_edit'].text()),
                        'num_steps': int(self.widget_export[trans]['steps_edit'].text()),
                        'function': self.checked_transformations[trans]['function']
                        }
                name = self.widget_export[trans]['name'].text()
                checked_transforms[name] = data

        export_images = plot_utils.get_all_single_transform_params(
            checked_transforms, num_steps='from_dict')

        # remove any params with value 0
        export_images = [x for x in export_images if not x[list(x.keys())[0]] == 0]

        # make and save images
        for single_trans in export_images:
            trans_name = list(single_trans.keys())[0]
            param = single_trans[trans_name]
            img = image_utils.get_transform_image(self.data_stores[0],
                                                  transform_functions={trans_name: self.checked_transformations[trans_name]},
                                                  transform_params={trans_name: param})
            trans_info = {'transform_name': trans_name, 'transform_value': param}
            image_utils.save_image(img, os.path.join(
                save_folder, f'{save_utils.make_name_for_trans(trans_info)}.png'))
        self.status_bar.showMessage(f'Images saved to {save_folder}', 5000)
        self.export_in_progress = False
        self.last_export = save_folder



    def open_mlp_new(self, mpl_canvas):
        fig = gui_utils.break_out_mlp(mpl_canvas)
        fig.show()

    '''
    experiments
    '''
    def change_text_exp_trans(self, trans):
        ''' change colour of text when checkbox is pressed '''
        checked = self.widget_experiment_params[trans]['check_box'].isChecked()
        for widget in self.widget_experiment_params[trans]:
            if widget == 'check_box':
                pass
            else:
                self.widget_experiment_params[trans][widget].setEnabled(checked)
                if checked == True:
                    self.widget_experiment_params[trans][widget].setStyleSheet(f"QLineEdit {{color: {self.settings_text_colour_original};}}\nQLabel {{color: {self.settings_text_colour_original};}}")
                else:
                    self.widget_experiment_params[trans][widget].setStyleSheet(f"QLineEdit {{color: gray;}}\nQLabel {{color: gray;}}")

    def launch_experiment_JND(self):
        '''launch the Just Noticable difference experiment'''
        # first get all the checked transforms and their parameters
        checked_transformation_params = {}
        for trans in self.widget_experiment_params:
            if self.widget_experiment_params[trans]['check_box'].isChecked():
                data = {'min': make_float_from_text(self.widget_experiment_params[trans]['min_edit'].text()),
                        'max': make_float_from_text(self.widget_experiment_params[trans]['max_edit'].text()),
                        'num_steps': int(self.widget_experiment_params[trans]['steps_edit'].text()),
                        'function': self.checked_transformations[trans]['function']}
                name = self.widget_experiment_params[trans]['name'].text()
                checked_transformation_params[name] = data

        # check that we only have one transform selected to JND test to work
        if len(checked_transformation_params) != 1:
            # show message to UI if more than one transform
            self.status_bar.showMessage(
                '!!! Select only one tranform/distortion for Just Noticable Difference Experiment !!!', 10000)
        else:
            # run JND experiment
            self.experiment_JND = IQM_Vis.UI.make_experiment_JND(checked_transformation_params,
                                                                 self.data_stores[0],
                                                                 self.image_display_size,
                                                                 self.rgb_brightness,
                                                                 self.display_brightness,
                                                                 self.default_save_dir,
                                                                 self.default_dataset_name,
                                                                 self.pre_processing_option,
                                                                 self.post_processing_option,
                                                                 self.exp_range_lower,
                                                                 self.exp_range_upper,
                                                                 self.checked_metrics)
            self.experiment_JND.saved_experiment.connect(
                self.change_human_scores_after_exp_JND)
            self.experiment_JND.show()
            self.experiment_JND.showFullScreen()
            

    def launch_experiment_2AFC(self):
        '''Launch the 2 alternate forced choice experiment'''
        # first get all the checked transforms and their parameters
        checked_transformation_params = {}
        for trans in self.widget_experiment_params:
            if self.widget_experiment_params[trans]['check_box'].isChecked():
                data = {'min': make_float_from_text(self.widget_experiment_params[trans]['min_edit'].text()),
                        'max': make_float_from_text(self.widget_experiment_params[trans]['max_edit'].text()),
                        'num_steps': int(self.widget_experiment_params[trans]['steps_edit'].text()),
                        'function': self.checked_transformations[trans]['function']}
                name = self.widget_experiment_params[trans]['name'].text()
                checked_transformation_params[name] = data
        if self.checked_transformations != {}:
            self.experiment_2AFC = IQM_Vis.UI.make_experiment_2AFC(checked_transformation_params,
                                                         self.data_stores[0],
                                                         self.image_display_size,
                                                         self.rgb_brightness,
                                                         self.display_brightness,
                                                         self.default_save_dir,
                                                         self.pre_processing_option,
                                                         self.post_processing_option,
                                                         self.checked_metrics)
            self.experiment_2AFC.saved_experiment.connect(
                self.change_human_scores_after_exp_2AFC)
            self.experiment_2AFC.show()
            self.experiment_2AFC.showFullScreen()
        else:
            self.status_bar.showMessage('Cannot run experiment without transforms', 5000)

    @pyqtSlot(str)
    def change_human_scores_after_exp_2AFC(self, path):
        self._change_human_exp_2AFC(path)

    @pyqtSlot(str)
    def change_human_scores_after_exp_JND(self, path):
        self._change_human_exp_JND(path)


'''
float checking functionality for checkign line edit
'''

def get_float_validator():
    # float_validator = QDoubleValidator()
    # float_validator.setLocale(QLocale('en_US'))  # so comma cannot be used a decimal place for spanish users
    # float_validator.setNotation(QDoubleValidator.Notation(0)) # StandardNotation
    # return float_validator
    return custom_float_validator()

def is_float(input_string):
    # if input_string == '-' or input_string == '+' or input_string == '':
    #     return False
    # pattern = r'^[-+]?[0-9]*\.?[0-9]*$'
    # match = re.match(pattern, input_string)
    # if match:
    #     return True
    # return False
    try: 
        float(input_string)
        return True
    except ValueError:
        return False

def is_almost_float(input_string):
    # if input_string == '-' or input_string == '+' or input_string == '':
    #     return True
    # pattern = r'^[-+]?[0-9]*.*[0-9]$'
    # match = re.match(pattern, input_string)
    # if match:
    #     return True
    # return False
    return True

class custom_float_validator(QValidator):
    def __init__(self, parent=None):
        super(custom_float_validator, self).__init__(parent)

    def validate(self, string, pos):
        if is_float(string):
            return QValidator.State(2), string, pos # Acceptable
        if is_almost_float(string):
            return QValidator.State(1), string, pos # Intermediate
        return QValidator.State(0), string, pos # Invalid

def make_float_from_text(txt):
    # get rid of pesky commas
    no_commas = txt.replace(',', '')
    return float(no_commas)


'''
utils for saving experiments, images and figures
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
import json
import pickle
import pandas as pd

from IQM_Vis.utils import image_utils

DEFAULT_SAVE_DIR = os.path.join(os.path.expanduser("~"), 'IQM-Vis')

''' getters for experiment files from save dir '''
def get_human_scores_file(dir):
    last_dir = os.path.basename(os.path.normpath(dir))
    prefix = last_dir.split('-')[0]
    return os.path.join(dir, f'{prefix}-HUMAN-scores.csv')

def get_image_name_from_human_scores(dir):
    file_path = get_human_scores_file(dir)
    return os.path.basename(file_path)[:-len('-HUMAN-scores.csv')]

def get_human_times_file(dir):
    last_dir = os.path.basename(os.path.normpath(dir))
    prefix = last_dir.split('-')[0]
    return os.path.join(dir, f'{prefix}-HUMAN-times-taken.csv')

def get_IQM_file(dir):
    last_dir = os.path.basename(os.path.normpath(dir))
    prefix = last_dir.split('-')[0]
    return os.path.join(dir, f'{prefix}-IQM-scores.csv')

def get_original_image_file(dir):
    return os.path.join(dir, 'reference.png')

def get_original_unprocessed_image_file(dir):
    last_dir = os.path.basename(os.path.normpath(dir))
    prefix = last_dir.split('-')[0]
    return os.path.join(dir, f'{prefix}.png')

def get_image_processing_file(dir):
    return os.path.join(dir, 'transforms', 'processing.json')

def get_transform_params_file(dir):
    return os.path.join(dir, 'transforms', 'transform_params.pkl')

def get_transform_functions_file(dir):
    return os.path.join(dir, 'transforms', 'transform_functions.pkl')

''' other save helpers '''
def save_obj(pickle_path, trans):
    ''' save transforms as pickle file '''
    with open(pickle_path, 'wb') as file:
        pickle.dump(trans, file)

def load_obj(pickle_path):
    ''' load trans dict from pkl file '''
    if not os.path.exists(pickle_path):
        return None
    try:
        with open(pickle_path, 'rb') as file:
            trans = pickle.load(file)
        return trans
    except:
        return None
    
def get_JND_image_names(dir):
    ''' load the image names from a JND experiment '''
    images_dir = get_JND_ref_image_dir(dir)
    if not os.path.exists(images_dir):
        return None
    try:
        # get all file names in the directory
        return os.listdir(images_dir)
    except:
        return None
    
def get_JND_ref_image_dir(dir):
    return  os.path.join(dir, 'reference_images')

def get_JND_ref_image_unprocessed_dir(dir):
    return  os.path.join(dir, 'reference_images_unprocessed')


def save_df_as_csv(df, file, index=False):
    if os.path.exists(file):
        df_saved = pd.read_csv(file)
        df = pd.concat([df_saved, df])
    df.to_csv(file, index=index)


def save_and_merge_df_as_csv(df, file):
    '''df need to be indexed'''
    if os.path.exists(file):
        df_saved = pd.read_csv(file)
        index_name = df.index.name
        df = pd.merge(df.reset_index(), df_saved, how='outer')
        df.set_index(index_name, inplace=True)
    df.to_csv(file, index=True)


def save_and_merge_rm_duplicates_df_as_csv(df, file):
    '''save only new IQM rows'''
    if os.path.exists(file):
        df_saved = pd.read_csv(file)
        index_name = df.index.name
        df = pd.merge(df.reset_index(), df_saved, how='outer')
        df.set_index(index_name, inplace=True)
        # remove duplicates
        df = df[~df.index.duplicated(keep='last')]
    # sort columns for consistency
    cols = df.columns.tolist()
    cols.sort()
    df.to_csv(file, index=True, columns=cols)


def get_JND_user_ID(dir):
    human_experiment_csv_file = get_human_scores_file(dir)
    if not os.path.exists(human_experiment_csv_file):
        return 1
    df = pd.read_csv(human_experiment_csv_file)
    # get the last user ID
    return df['User ID'].max() + 1

def save_JND_experiment_results(experiment_results,
                                save_dir,
                                IQM_scores_df=None):

    # User ID
    ID = get_JND_user_ID(save_dir)
    results = {'ImageName': [],
               'UserDecision': [],
               'Transform': [],
               'TimeTaken': [],
               'User ID': []}
    for image_data in experiment_results:
        results['ImageName'].append(image_data['ref_name'])
        results['UserDecision'].append(image_data['user_decision'])
        results['Transform'].append(make_name_for_trans(image_data))
        results['TimeTaken'].append(image_data['time_taken'])
        results['User ID'].append(ID)
        # results[make_name_for_trans(image_data)] = [image_data['user_decision']]

    # make df
    df = pd.DataFrame.from_dict(results)
    # save results
    human_experiment_csv_file = get_human_scores_file(save_dir)
    save_df_as_csv(df, human_experiment_csv_file, index=False)

    # save IQM results
    if not isinstance(IQM_scores_df, type(None)) and not IQM_scores_df.empty:
        IQM_file = get_IQM_file(save_dir)
        save_and_merge_rm_duplicates_df_as_csv(IQM_scores_df, IQM_file)

    return human_experiment_csv_file

    
def save_2AFC_experiment_results(trans_names, 
                                results_order, 
                                save_dir, 
                                times_taken=None, 
                                IQM_scores_df=None):
    '''save all the experiment reults as csvs'''
    # make HIQM scores from ordering : HIQM = pos/num_pos
    results = {}
    for i, trans_name in enumerate(results_order):
        results[str(trans_name)] = [(i+1) / len(results_order)]

    # make df
    df = pd.DataFrame.from_dict(results)
    # re order so that the trans_names are acending order as given
    df = df[trans_names]
    # save results
    human_experiment_csv_file = get_human_scores_file(save_dir)
    save_df_as_csv(df, human_experiment_csv_file, index=False)

    # save times taken
    if times_taken != None:
        times_file = get_human_times_file(save_dir)
        data = {'mean time (seconds)': [sum(times_taken)/len(times_taken)],
                'total time (seconds)': [sum(times_taken)],
                'number of comparisons': [len(times_taken)],
                'individual times (seconds)': [str(times_taken)]}
        save_df_as_csv(pd.DataFrame.from_dict(data), times_file)

    # save IQM results
    if not isinstance(IQM_scores_df, type(None)) and not IQM_scores_df.empty:
        print(IQM_scores_df)
        IQM_file = get_IQM_file(save_dir)
        save_and_merge_rm_duplicates_df_as_csv(IQM_scores_df, IQM_file)

    return human_experiment_csv_file

def save_json_dict(path, dict_):
    with open(path, 'w') as fp:
        json.dump(dict_, fp)

def load_json_dict(path):
    with open(path, 'r') as fp:
        return json.load(fp)

def make_name_for_trans(trans):
    # splitter = '-----'
    splitter = '<>'
    return f"{trans['transform_name']}{splitter}{trans['transform_value']}"



'''
matplotlib plotting helpers
TODO: write dev_resources/docs how to use these (currently just have to look at the UI code)
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from functools import partial
import math

import numpy as np
import scipy.stats

import IQM_Vis
from IQM_Vis.utils import image_utils, gui_utils

'''
plot bar chart on matplotlib qt qidget
'''
class bar_plotter:
    def __init__(self, bar_names, var_names, ax, lim):
            self.bar_names = bar_names
            self.var_names = var_names
            self.ax = ax
            self.lim =lim
            self.ax.axes.clear()
            self.num_bars = len(self.bar_names)
            self.num_vars = len(self.var_names)
            self.bar_width = 1/(self.num_bars+1)
            self.bars = [np.arange(self.num_vars)]
            for i in range(1, self.num_bars):
                self.bars.append([x + self.bar_width for x in self.bars[i-1]])

    def plot(self, bar_name, var_values):
        if len(self.bar_names) > 1:
            i = self.bar_names.index(bar_name)
            self.ax.axes.bar(self.bars[i], var_values, width=self.bar_width, label=bar_name)
        else:
            self.ax.axes.bar(self.var_names, var_values)

    def show(self):
        self.set_style()
        self.ax.draw()

    def set_style(self):
        if len(self.bar_names) > 1:
            self.ax.axes.legend()
            self.ax.axes.set_xticks([r + self.bar_width for r in range(self.num_vars)], self.var_names)
        self.set_plot_lims()
        self.ax.figure.tight_layout()

    def set_plot_lims(self):
        y_lims = self.ax.axes.get_ylim()
        # self.ax.axes.set_ylim(min(0, y_lims[0], max(1, y_lims[1])))
        self.ax.axes.set_ylim(top= max(self.lim, y_lims[1]))

'''
line plot of matplotlib qt widget
'''
class line_plotter:
    def __init__(self, ax, x_label='', y_label='', lim=1):
        self.ax = ax
        self.ax.axes.clear()
        self.x_label = x_label
        self.y_label = y_label
        self.lim = lim

    def plot(self, x, y, label):
        self.ax.axes.plot(x, y, label=label)

    def show(self):
        self.set_style()
        self.ax.draw()

    def set_style(self):
        self.ax.axes.legend()
        self.ax.axes.set_xlabel(self.x_label)
        self.set_plot_lims()
        self.ax.figure.tight_layout()

    def set_plot_lims(self):
        y_lims = self.ax.axes.get_ylim()
        # self.ax.axes.set_ylim(min(0, y_lims[0], max(1, y_lims[1])))
        self.ax.axes.set_ylim(top= max(self.lim, y_lims[1]))

'''
plot radar chart on matplotlib qt widget
'''
class radar_plotter:
    def __init__(self, radar_names, var_names, ax, lim=1):
            self.radar_names = radar_names
            self.var_names = var_names
            self.ax = ax
            self.lim = lim
            self.ax.axes.clear()
            self.num_radars = len(self.radar_names)
            self.num_vars = len(self.var_names)
            self.radar_angles = [n/float(self.num_vars)*2*np.pi for n in range(self.num_vars)]
            self.radar_angles += self.radar_angles[:1]   # circular plot closure

    def plot(self, radar_name, var_values):
        var_values.append(var_values[0])   # add start to end to close the circular plot
        p = self.ax.axes.plot(self.radar_angles, var_values, linewidth=1, linestyle='solid', label=radar_name)
        self.ax.axes.fill(self.radar_angles, var_values, color=p[0].get_color(), alpha=0.1)

    def show(self):
        self.set_style()
        self.ax.draw()

    def set_style(self):
        # self.ax.axes.legend(loc='upper right', bbox_to_anchor=(0.1, 0.1))
        self.ax.axes.legend(fontsize='x-small', loc='upper left')
        self.ax.axes.set_xticks(self.radar_angles[:-1], self.var_names)
        self.set_plot_lims()
        # self.ax.fig.tight_layout()

    def set_plot_lims(self):
        y_lims = self.ax.axes.get_ylim()
        # self.ax.axes.set_ylim(min(0, y_lims[0], max(1, y_lims[1])))
        self.ax.axes.set_ylim(top= max(self.lim, y_lims[1]))

'''
scatter plotter on matplotlib qt widget
'''
class scatter_plotter:
    def __init__(self, ax, x_label='', y_label='', lim=1):
        self.ax = ax
        self.ax.axes.clear()
        self.x_label = x_label
        self.y_label = y_label
        self.lim = lim

    def plot(self, x, y, annotations=None, error=None):
        self.sc = self.ax.axes.scatter(x, y, picker=True, color='blue')
        if annotations != None:
            self.annotations = annotations
        if error != None or error != []:
            self.ax.axes.errorbar(x, y, yerr=error, fmt="o", capsize=3, color='black')

    def show(self):
        self.set_style()
        self.ax.draw()

    def set_style(self):
        self.ax.axes.set_xlabel(self.x_label)
        self.ax.axes.set_ylabel(self.y_label)
        # self.set_plot_lims()
        self.ax.figure.tight_layout()

    def set_plot_lims(self):
        y_lims = self.ax.axes.get_ylim()
        self.ax.axes.set_ylim(top=max(self.lim, y_lims[1]))

'''
metric averaging functions to get metric values over a range of transformation
'''
def get_all_slider_values(transforms, num_steps=11):
    if num_steps == 1:
        raise ValueError(f'number of steps cannot be 1')
    if num_steps == 2:
        return [transforms['min'], transforms['max']]
    float_stabiliser = 100    # work around since computer can't represent e.g. 0.2 very well
    num_steps -= 1    # since we are using steps then appending stop value
    range = transforms['max'] - transforms['min']
    # make sure if it's an int then we dont return step values that are floats
    if isinstance(transforms['min'], int) and isinstance(transforms['max'], int):
        num_steps = min(range, num_steps)
    steps = range/num_steps
    if isinstance(transforms['min'], int) and isinstance(transforms['max'], int):
        steps = int(steps)
    values = np.arange(start=transforms['min']*float_stabiliser,
                       stop=transforms['max']*float_stabiliser, 
                       step=steps*float_stabiliser)
    values = list(values/float_stabiliser)
    if transforms['max'] not in values:
        values.append(transforms['max'])
    return values

def get_all_single_transform_params(transforms, num_steps=11):
    ''' get a list of all the individual transforms with a single parameter value
        useful when doing experiments to make a dataset '''
    list_of_single_trans = []
    for trans_name, trans_data in transforms.items():  # loop over all transformations
        if num_steps == 'from_dict':
            steps = trans_data['num_steps']
        else:
            steps = num_steps
        for val in get_all_slider_values(trans_data, num_steps=steps):
            list_of_single_trans.append({trans_name: val})

    return list_of_single_trans

def compute_metric_for_human_correlation(data_store, transforms, metric_params, trans_str_values, metric):
    scores = {}
    for trans_str in trans_str_values:
        trans, trans_value = gui_utils.get_trans_dict_from_str(trans_str)
        if trans in transforms.keys():
            single_trans = {trans: transforms[trans]}
        else:
            iqm_vis_trans = IQM_Vis.transforms.get_all_transforms()
            if trans in iqm_vis_trans.keys():
                single_trans = {trans: iqm_vis_trans[trans]}
            else:
                raise Exception(f"Cannot find transformation {trans} in provided transforms or any default transforms, please provide an implimentation of {trans} when contructing the UI.")
        single_param_dict = {trans: trans_value}
        trans_im = image_utils.get_transform_image(data_store, single_trans, single_param_dict) # initialse image
        metric_scores = data_store.get_metrics(trans_im, metric, **metric_params)
        scores[trans_str] = metric_scores[metric]
    return scores

def compute_metrics_over_range_single_trans(data_store, transforms, metric_params, metrics_to_use, pbar_signal=None, stop_flag=None, num_steps=11):
    ''' compute metrics over a range of trans

    Args:
        data_store: object containing metrics and image
        transforms (dict): containing trans functions and min/max/initial values

    Returns:
        results (dict): results of IQM values at each transform value across its
                        whole paramter rance for the reference image
    '''
    # compute all metrics over their range of params and get avg/std
    results = {}
    # initialise results
    for metric in metrics_to_use:
        results[metric] = {}
        for tran in transforms:
            results[metric][tran] = {'param_values': [], 'scores': []}

    # get bar info (the lazy way)
    if pbar_signal != None:
        len_loop = 0
        for i, curr_trans in enumerate(transforms):
            for trans_value in get_all_slider_values(transforms[curr_trans], num_steps=num_steps):
                len_loop += 1
    pbar_counter = 0

    # compute over all image transformations
    for curr_trans in transforms:  # loop over all transformations
        single_trans = {curr_trans: transforms[curr_trans]}
        all_param_values = get_all_slider_values(
            transforms[curr_trans], num_steps=num_steps)
        for trans_value in all_param_values:
            single_param_dict = {curr_trans: trans_value}
            trans_im = image_utils.get_transform_image(data_store, single_trans, single_param_dict) # initialse image
            metric_scores = data_store.get_metrics(trans_im, metrics_to_use, **metric_params)
            for metric in metric_scores:
                results[metric][curr_trans]['scores'].append(float(metric_scores[metric]))
                # and store the input trans values for plotting
                results[metric][curr_trans]['param_values'] = all_param_values

                # end if flag says to stop
                if stop_flag != None:
                    if stop_flag[0] == True:
                        # reset pbar
                        if pbar_signal != None:
                            pbar_signal.emit(0)
                        return

            # send signal to progress bar if provided
            if pbar_signal != None:
                pbar_counter += 1
                if pbar_counter == len_loop:
                    pbar_signal.emit(0)
                else:
                    pbar_signal.emit(int(((pbar_counter)/len_loop)*100))
    return results

def compute_metrics_over_range(data_store, transforms, transform_values, metric_params, metrics_to_use, pbar_signal=None, stop_flag=None, num_steps=11):
    '''
    compute metrics over a range of trans (when using non initial values for other transforms)
    currently this method is not being used and instead using the simpler compute_metrics_over_range_single_trans

    Args:
        data_store: object containing metrics and image
        transforms (dict): containing trans functions and min/max/initial values
        transform_values (dict): containing the fixed current transform parameter values

    Returns:
        results (dict): results of IQM values at each transform value across its
                        whole paramter rance for the reference image
    '''
    # compute all metrics over their range of params and get avg/std
    results = {}
    # initialise results
    for metric in data_store.metrics:
        results[metric] = {}
        for tran in transforms:
            results[metric][tran] = {'param_values': [], 'scores': []}

    # get bar info (the lazy way)
    if pbar_signal != None:
        len_loop = 0
        for i, curr_trans in enumerate(transforms):
            for trans_value in get_all_slider_values(transforms[curr_trans], num_steps=num_steps):
                len_loop += 1
    pbar_counter = 0

    # compute over all image transformations
    for curr_trans in transforms:  # loop over all transformations
        for trans_value in get_all_slider_values(transforms[curr_trans], num_steps=num_steps):   # all values of the parameter
            vary_one_value = transform_values.copy()
            vary_one_value[curr_trans] = trans_value   # set to the varying value in this range loop
            trans_im = image_utils.get_transform_image(data_store, transforms, vary_one_value)     # initialse image
            metric_scores = data_store.get_metrics(trans_im, metrics_to_use, **metric_params)
            for metric in metric_scores:
                results[metric][curr_trans]['scores'].append(float(metric_scores[metric]))
                # and store the input trans values for plotting
                results[metric][curr_trans]['param_values'] = get_all_slider_values(
                    transforms[curr_trans], num_steps=num_steps)

                # end if flag says to stop
                if stop_flag != None:
                    if stop_flag[0] == True:
                        # reset pbar
                        if pbar_signal != None:
                            pbar_signal.emit(0)
                        return

            # send signal to progress bar if provided
            if pbar_signal != None:
                pbar_counter += 1
                if pbar_counter == len_loop:
                    pbar_signal.emit(0)
                else:
                    pbar_signal.emit(int(((pbar_counter)/len_loop)*100))
    return results

def get_radar_plots_avg_plots(results, metrics_names, transformation_names, axes, lim=1):
    '''
    plot results on a polar axes -> radar/spider plot
    '''
    radar_plt = radar_plotter(radar_names=metrics_names,
                                    var_names=transformation_names,
                                    ax=axes,
                                    lim=lim)
    if transformation_names == []:
        return radar_plt
    for metric in metrics_names:
        mean_value = []
        # std_value = []
        transform = []
        for tran in transformation_names:
            transform.append(tran)
            mean_value.append(np.mean(results[metric][tran]['scores']))
            # std_value.append(np.std(results[metric][tran]['scores']))
        radar_plt.plot(metric, mean_value)
    radar_plt.set_style()
    return radar_plt

def get_transform_range_plots(results, transform, axes, lim=1):
    '''
    plot a single transform range graph of all metrics
    '''
    plot = line_plotter(axes, transform, 'Values', lim=lim)
    for metric in results:
        plot.plot(results[metric][transform]['param_values'], results[metric][transform]['scores'], metric)
    return plot

def get_correlation_plot(human_scores, metric_scores, axes, metric, change_trans_value_signal):
    '''
    scatter plot for correlations
    '''
    sp = scatter_plotter(axes,
                x_label=metric,
                y_label='Human Score')
    x = []
    y = []
    e = []
    labels = []
    for trans in human_scores['mean']:
        x.append(metric_scores[metric][trans])
        y.append(human_scores['mean'][trans])
        try:
            std = human_scores['std'][trans]
        except KeyError:
            std = 0
        if math.isnan(std):
            std = 0
        e.append(std)
        name, value = gui_utils.get_trans_dict_from_str(trans)
        labels.append(f'{name}={value}')
    sp.plot(x, y, annotations=labels, error=e)
    # make interactive hover for points
    annot = sp.ax.axes.annotate("", xy=(0, 0), xytext=(0, 0), textcoords="offset points",
                        bbox=dict(boxstyle="round", fc="w"),
                        arrowprops=dict(arrowstyle="->")
                        )
    annot.set_visible(False)
    sp.ax.figure.canvas.mpl_connect(
        "motion_notify_event", partial(hover_scatter, sp, annot))
    sp.ax.figure.canvas.mpl_connect(
        "pick_event", partial(click_scatter, sp, change_trans_value_signal))
    # get correlation as title
    pear = scipy.stats.pearsonr(x, y)
    spear = scipy.stats.spearmanr(x, y)
    sp.ax.axes.set_title(
        f"Spearman's: {spear.correlation:.4f}\n  Pearson's: {pear.statistic:.4f}")
    return sp

def click_scatter(_plot, change_trans_value_signal, event):
    ''' send signal of which data point was clicked '''
    change_trans_value_signal.emit(_plot.annotations[event.ind[0]])

def hover_scatter(_plot, annot, event):
    vis = annot.get_visible()
    # if event.inaxes == _plot.ax.axes:
    cont, ind = _plot.sc.contains(event)
    if cont:
        update_annot(ind, _plot, annot)
        annot.set_visible(True)
        _plot.ax.figure.canvas.draw_idle()
    else:
        if vis:
            annot.set_visible(False)
            _plot.ax.figure.canvas.draw_idle()

def update_annot(ind, _plot, annot):
    pos = _plot.sc.get_offsets()[ind["ind"][0]].copy()
    xlims = _plot.ax.axes.get_xlim()
    x_range = xlims[1] - xlims[0]
    x_middle = xlims[0] + x_range/2
    if pos[0] > x_middle:
        pos[0] -= x_range/2
    # else:
    #     pos[0] += xlim/5
    annot.xy = pos
    text = ""
    for i in ind['ind']:
        if text != "":
            text += "\n"
        text += f"{_plot.annotations[i]}"
    annot.set_text(text)



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from IQM_Vis.metrics.non_perceptual import MAE, MSE, RMSE, one_over_PSNR
from IQM_Vis.metrics.perceptual_trad import SSIM, MS_SSIM, NLPD
from IQM_Vis.metrics.perceptual_DL import LPIPS, DISTS

def get_all_metrics():
    ''' 
    Get all available IQMs provided by IQM_Vis 
    
    Return:
        - all_metrics (dict)
        '''
    all_metrics = {
        'MSE': MSE(),
        'NLPD': NLPD(),
        '1-SSIM': SSIM(),
        '1-MS_SSIM': MS_SSIM(),
        '1/PSNR': one_over_PSNR(),
        'DISTS': DISTS(),
        'LPIPS': LPIPS(),
        'MAE': MAE(),
        'RMSE': RMSE(),
    }
    return all_metrics

def get_all_metric_images():
    ''' 
    Get all available IQMs provided by IQM_Vis that return an image 
    
    Return:
        - all_metrics (dict)
    '''
    all_metrics = {
    'MSE': MSE(return_image=True),
    'SSIM': SSIM(return_image=True),
    # 'MS_SSIM': MS_SSIM(return_image=True),
    'MAE': MAE(return_image=True),
    'RMSE': RMSE(return_image=True),
    }
    return all_metrics

def get_all_IQM_params():
    ''' 
    Get all available IQMs parameters
    
    Return:
        - all_params (dict)
    '''
    all_params = {
        'sigma': {'min':0.25, 'max':5.25, 'init_value': 1.5},  # for the guassian kernel
        'k1': {'min':0.01, 'max':0.21, 'init_value': 0.01},
        'k2': {'min':0.01, 'max':0.21, 'init_value': 0.03},
        'ssim_kernel_size': {'min':3, 'max':21, 'normalise':'odd', 'init_value': 11},  # ignored if guassian kernel used
        'mssim_kernel_size': {'min':3, 'max':21, 'normalise':'odd', 'init_value': 11},  # ignored if guassian kernel used
    }
    return all_params


'''
blueprint for data loader API
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from abc import ABC, abstractmethod

class base_dataloader(ABC):
    @property
    def metrics(self):
        return self._metrics

    @metrics.setter
    def metrics(self, metrics_dict):
        self._metrics = metrics_dict

    @property
    def metric_images(self):
        return self._metric_images

    @metric_images.setter
    def metric_images(self, metric_images_dict):
        self._metric_images = metric_images_dict

    @abstractmethod
    def get_reference_image_name(self):
        pass

    @abstractmethod
    def get_reference_image(self):
        pass

    @abstractmethod
    def get_image_to_transform_name(self):
        pass

    @abstractmethod
    def get_image_to_transform(self):
        pass

    @abstractmethod
    def get_metrics(self):
        pass

    @abstractmethod
    def get_metric_images(self):
        pass


class base_dataset_loader(base_dataloader):
    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __getitem__(self):
        pass

    @abstractmethod
    def get_reference_image_by_index(self, index):
        pass



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License
import warnings
import torch

from torchmetrics.image.lpip import LearnedPerceptualImagePatchSimilarity as lpips_torch
from DISTS_pytorch import DISTS as dists_original
from IQM_Vis.metrics.metric_utils import _check_shapes, _numpy_to_torch_image


class LPIPS:
    '''Learned Perceptual Image Patch Similarity between two images.
       Images must have the same dimensions.

    Args:
        network (str): Pretrained network to use. Choose between ‘alex’, ‘vgg’
                       or ‘squeeze’. (Defaults to 'alex')
        reduction (str): How to reduce over the batch dimension. Choose between
                         ‘sum’ or ‘mean’. (Defaults to 'mean')

    '''

    def __init__(self, network='alex', reduction='mean'):
        self.initialised = False   # initialse fully on first __call__ to save load up time
        self.network = network
        self.reduction = reduction
        # self.metric = lpips_torch
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu")
        self.preproccess_function = _numpy_to_torch_image

    def __call__(self, im_ref, im_comp, **kwargs):
        '''When an instance is called

        Args:
            im_ref (np.array): Reference image
            im_comp (np.array): Comparison image
            **kwargs: Arbitrary keyword arguments

        Returns:
            score (np.array): LPIPS score
        '''
        if self.initialised == False:
            with warnings.catch_warnings():    # we don't care about the warnings these give
                warnings.simplefilter("ignore")
                self.metric = lpips_torch(net_type=self.network,
                                          reduction=self.reduction,
                                          normalize=True)
                self.metric.to(self.device)
            self.initialised = True
        _check_shapes(im_ref, im_comp)
        im_ref = self.preproccess_function(im_ref).to(
            device=self.device, dtype=torch.float)
        im_comp = self.preproccess_function(im_comp).to(
            device=self.device, dtype=torch.float)
        with warnings.catch_warnings():    # warning because we have reset
            warnings.simplefilter("ignore")
            _score = self.metric(im_ref, im_comp)
        score = _score.cpu().detach().numpy()
        self.metric.reset()
        return score


class DISTS:
    '''Deep Image Structure and Texture Similarity (DISTS) Metric. Uses the
        code from https://github.com/dingkeyan93/DISTS. Uses the PyTorch backend.
        It is robust to texture variance (e.g., evaluating the images generated
        by GANs) and mild geometric transformations (e.g., evaluating the image
        pairs that are not strictly point-by-point aligned).

    '''

    def __init__(self):
        self.initialised = False   # initialse fully on first __call__ to save load up time
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu")
        self.preproccess_function = _numpy_to_torch_image

    def __call__(self, im_ref, im_comp, **kwargs):
        '''When an instance is called

        Args:
            im_ref (np.array): Reference image
            im_comp (np.array): Comparison image
            **kwargs: Arbitrary keyword arguments

        Returns:
            score (np.array): DISTS score
        '''
        # load model on first time called
        if self.initialised == False:
            with warnings.catch_warnings():    # we don't care about the warnings these give
                warnings.simplefilter("ignore")
                self.metric = dists_original()
            self.metric.to(self.device)
            self.initialised = True

        _check_shapes(im_ref, im_comp)
        im_ref = self.preproccess_function(im_ref).to(
            device=self.device, dtype=torch.float)
        im_comp = self.preproccess_function(im_comp).to(
            device=self.device, dtype=torch.float)
        _score = self.metric(im_ref, im_comp)
        score = _score.cpu().detach().numpy()
        return score



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License
import pytest
from PyQt6 import QtTest, QtWidgets, QtCore
from pytestqt.plugin import QtBot
import IQM_Vis


def default_UI():
    app = IQM_Vis.make_UI(test=True)
    return app


# building and closing function of UI for testing
class BotTester:
    def __init__(self, get_UI=default_UI, wait_time=100, final_wait=False):
        self.get_UI = get_UI
        self.wait_time = wait_time
        self.final_wait = final_wait

    @pytest.fixture(scope='function')
    def build_IQM_Vis(self):
        # Setup
        test_window = self.get_UI()
        qtbotbis = QtBot(test_window.window)

        yield test_window, qtbotbis

        # Clean up
        QtTest.QTest.qWait(1000)

        # need to handle the closing dialog
        def handle_dialog():
            messagebox = QtWidgets.QApplication.activeWindow()
            yes_button = messagebox.button(
                QtWidgets.QMessageBox.StandardButton.Yes)
            qtbotbis.mouseClick(
                yes_button, QtCore.Qt.MouseButton.LeftButton, delay=1)

        QtCore.QTimer.singleShot(100, handle_dialog)

        test_window.window.main.close()
        # if self.final_wait:
        QtTest.QTest.qWait(1000)






'''
API to access making the PyQt6 UI for IQM-Vis
TODO: write dev_resources/docs on example usage/ what inputs etc. and what attributes that the data_store class needs
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import subprocess
import platform
import sys
from typing import Any
import warnings
import types
import numpy as np

try:
    from PyQt6.QtWidgets import QApplication
    import IQM_Vis
    from IQM_Vis.utils import image_utils
    import matplotlib
    matplotlib.use("Qt5Agg")
except ImportError:
    warnings.warn('Cannot load PyQt6 library - running IQM_Vis package in headless mode')
    
from IQM_Vis.UI.main import make_app
from IQM_Vis.examples.images import DEFAULT_IMAGES


class make_UI:
    def __init__(self, 
                 data_store=None,
                 transformations=None,
                 image_list: list=DEFAULT_IMAGES,
                 metrics: dict={},
                 metric_images: dict={},
                 metrics_info_format: str='graph',
                 metrics_avg_graph: bool=True,
                 metric_params: dict={},
                 default_save_dir=IQM_Vis.utils.save_utils.DEFAULT_SAVE_DIR,
                 default_dataset_name='dataset1',
                 restrict_options=None,
                 num_steps_range=11,
                 debug=False, 
                 test=False):
        if data_store == None:
            data_store = IQM_Vis.dataset_holder(image_list,
                                   metrics,
                                   metric_images)
        self.data_store = data_store
        self.transformations = transformations
        self.metrics_info_format = metrics_info_format
        self.metrics_avg_graph = metrics_avg_graph
        self.metric_params = metric_params
        self.default_save_dir = default_save_dir
        self.default_dataset_name = default_dataset_name
        self.restrict_options = restrict_options
        self.num_steps_range = num_steps_range
        self.debug = debug
        self.test = test
        check_pyqt_install_deps()
        self.show()
        self.showing = True

    def show(self):
        self._check_restrict_options()
        self._check_data_store()
        self._check_trans()
        if self.debug:
            self._check_inputs()
        self.app = QApplication(sys.argv)
        self.window = make_app(self.app,
                               self.data_store,
                               self.transformations,
                               metrics_info_format=self.metrics_info_format,
                               metrics_avg_graph=self.metrics_avg_graph,
                               metric_params=self.metric_params,
                               default_save_dir=self.default_save_dir,
                               default_dataset_name=self.default_dataset_name,
                               restrict_options=self.restrict_options,
                               num_steps_range=self.num_steps_range,
                               test=self.test)
        if self.test == False:
            sys.exit(self.app.exec())

    def _check_restrict_options(self):
        if self.restrict_options == None:
            trans = len(self.transformations) if self.transformations != None else 0
            if isinstance(self.data_store, list):
                metrics = len(self.data_store[0].metrics)
                metric_images = len(self.data_store[0].metric_images)
            else:
                metrics = len(self.data_store.metrics) if self.data_store != None else 0
                metric_images = len(self.data_store.metric_images) if self.data_store != None else 0

            self.restrict_options = {'transforms': trans,
                                     'metrics': metrics,
                                     'metric_images': metric_images}
        if isinstance(self.restrict_options, int):
            self.restrict_options = {'transforms': self.restrict_options,
                                     'metrics': self.restrict_options,
                                     'metric_images': self.restrict_options}



    def _check_data_store(self):
        '''data store checking'''
        if self.data_store == None:
            self.data_store = IQM_Vis.dataset_holder(IQM_Vis.examples.images.DEFAULT_IMAGES,
                IQM_Vis.metrics.get_all_metrics(),
                IQM_Vis.metrics.get_all_metric_images()
                )
            if self.metric_params == {}:
                self.metric_params = IQM_Vis.metrics.get_all_IQM_params()
        if not isinstance(self.data_store, list):
            self.data_store = [self.data_store]

    def _check_trans(self):
        if self.transformations == None:
            self.transformations = IQM_Vis.transforms.get_all_transforms()
        # make sure to wrap all transforms in clip so they don't go beyond data limits
        for trans, data in self.transformations.items():
            self.transformations[trans]['function'] = transform_wrapper(data['function'])

    def _check_inputs(self):
        for item in self.data_store:
            test_datastore_attributes(item)
        '''input items that should be dictionaries'''
        should_be_dict = [self.transformations, self.metric_params]
        for item in should_be_dict:
            if not isinstance(item, dict):
                var_name = f'{item=}'.split('=')[0]
                raise TypeError('make_UI input: '+var_name+' should be a dictionary not '+str(type(item)))
            # elif len(item.keys()) == 0:
            #     var_name = f'{item=}'.split('=')[0]
            #     warnings.warn(f'make_UI input: {var_name} is empty')

        '''extra input option checks'''
        if type(self.metrics_info_format) != str:
            var_name = f'{self.metrics_info_format=}'.split('=')[0]
            raise TypeError('make_UI input: '+var_name+' should be a string not '+str(type(self.metrics_info_format)))
        if type(self.metrics_avg_graph) != bool:
            var_name = f'{self.metrics_info_format=}'.split('=')[0]
            raise TypeError('make_UI input: '+var_name+' should be a bool not '+str(type(self.metrics_info_format)))


def test_datastore_attributes(data_store):
    '''get the data handler class to make sure its properties are correct'''
    obj_name = data_store.__class__.__name__
    attributes = [
    ('get_reference_image', 'gets the input reference image', types.MethodType),
    ('get_reference_image_name', 'gets the name of the input reference image', types.MethodType),
    ('get_image_to_transform', 'gets the input image to transform', types.MethodType),
    ('get_image_to_transform_name', 'gets the name of the input image to transform', types.MethodType),
    ('metrics', 'is a dict of all metric functions', dict),
    ('metric_images','is a dict of all metric image functions', dict),
    ('get_metrics', 'is a function that returns a dict of all metrics results', types.MethodType),
    ('get_metric_images', 'is a function that returns a dict of all metric image results', types.MethodType),
    ]
    for att in attributes:
        if hasattr(data_store,  att[0]):
            attr = getattr(data_store,  att[0])
            if type(attr) != att[2]:
                raise TypeError(f"{obj_name} attribute '{att[0]}' needs to be type '{att[2]}' instead of {type(attr)}")
        else:
            raise AttributeError(f"{obj_name} needs to have attribute '{att[0]}' which {att[1]}")

    # now test returned types are correct
    method_return_types = [
    ('get_reference_image', np.ndarray),
    ('get_reference_image_name', str),
    ('get_reference_image', np.ndarray),
    ('get_image_to_transform_name', str),
    ('get_metrics', dict, image_utils.get_transform_image(data_store, {}, {})),
    ('get_metric_images', dict, image_utils.get_transform_image(data_store, {}, {})),
    ]
    for meth in method_return_types:
        method = getattr(data_store,  meth[0])
        if len(meth) < 3:
            ret = method()
        elif len(meth) == 3:
            ret = method(meth[2])
        if type(ret) != meth[1]:
            raise TypeError(f"{obj_name} method '{meth[0]}' needs to return type '{att[2]}' instead of {type(attr)}")


class transform_wrapper:
    ''' wrap transforms to make sure they return an image between [0, 1] '''
    def __init__(self, function):
        self.function = function

    def __call__(self, *args, **kwargs):
        called = self.function(*args, **kwargs)
        return np.clip(called, 0, 1)
    
    def __eq__(self, other):
        if isinstance(other, transform_wrapper):
            return self.function == other.function
        else:
            # unwrapped tranform function
            return self.function == other


def check_pyqt_install_deps():
    ''' PyQt6 on linux doesn't always ship with all the required libraries
        so here we will warn the user if they need to install them, avoiding a crash '''
    system_info = platform.system()
    if not system_info == "Linux":
        return True
    
    # if on linux try see if package installed (only for debian based)
    package_name = "libxcb-cursor0"
    installed = False
    try:
        result = subprocess.run(["dpkg", "-l", package_name], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        installed = "ii" in result.stdout
    except subprocess.CalledProcessError:
        installed = False
    if installed == False:
        print(f"{'*'*30}\n\nWarning: not all dependencies are installed. If you get an 'Aborted (core dumped)' error below this message then please install them using:\n\n 'sudo apt install libxcb-cursor0'\n\n\n{'*'*30}\n\n")



from IQM_Vis.UI.images import images
from IQM_Vis.UI.widgets import widgets
from IQM_Vis.UI.layout import layout
from IQM_Vis.UI.custom_widgets import ProgressBar, ClickLabel
from IQM_Vis.UI.threads import get_range_results_worker
from IQM_Vis.UI.experiment_mode_2AFC import make_experiment_2AFC
from IQM_Vis.UI.experiment_mode_JND import make_experiment_JND
